# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""MDODiscipline builder from a python function."""

from __future__ import division, unicode_literals

import logging
import re
from inspect import getsource

from numpy import array, atleast_2d, ndarray

from gemseo.core.data_processor import DataProcessor
from gemseo.core.discipline import MDODiscipline
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.py23_compat import getargspec

LOGGER = logging.getLogger(__name__)


class AutoPyDiscipline(MDODiscipline):
    """A simplified and straightforward way of integrating a discipline from a python
    function.

    Examples
    --------
    >>> from gemseo.core.auto_py_discipline import AutoPyDiscipline
    >>> from numpy import array
    >>> def my_function(x=0., y=0.):
    >>>     z = x + 2*y
    >>>     return z
    >>> discipline = AutoPyDiscipline(py_func=my_function)
    >>> discipline.execute()
    {'x': array([0.]), 'y': array([0.]), u'z': array([0.])}
    >>> discipline.execute({'x': array([1.]), 'y':array([-3.2])})
    {'x': array([1.]), 'y': array([-3.2]), u'z': array([-5.4])}

    There are a few constraints:

    - only one return statement,
    - return must return a variable reference or a list of references,
    - only floats or arrays as inputs and outputs.

    See also
    --------
    gemseo.core.discipline.MDODiscipline : abstract class defining
        the key concept of discipline
    """

    def __init__(self, py_func, py_jac=None, use_arrays=False, write_schema=False):
        """Constructor.

        :param py_func: the python function to be used to generate the
            MDODiscipline.
        :type py_func: function
        :param use_arrays: if True, the function is expected to take arrays
            as inputs and give outputs as arrays.
        :type use_arrays: bool
        :param py_jac: pointer to the function jacobian;
            the jacobian must be a 2D numpy array.
        :type py_jac: function
        :param write_schema: if True, write the json schema on the disk
        :type write_schema: bool
        """
        if not callable(py_func):
            raise TypeError("py_func must be callable!")

        super(AutoPyDiscipline, self).__init__(
            name=py_func.__name__,
            auto_detect_grammar_files=False,
            grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        )

        self.py_func = py_func
        self.use_arrays = use_arrays
        self.py_jac = py_jac

        args_in = getargspec(py_func)[0]  # pylint: disable=deprecated-method
        self.in_names = args_in
        self.input_grammar.initialize_from_data_names(self.in_names)
        self.out_names = self._get_return_spec(py_func)
        self.output_grammar.initialize_from_data_names(self.out_names)

        if write_schema:
            self.input_grammar.write_schema()
            self.output_grammar.write_schema()

        if not use_arrays:
            self.data_processor = AutoDiscDataProcessor(self.out_names)

        if self.py_jac is None:
            self.set_jacobian_approximation()

        def_func = self._get_defaults()
        self.default_inputs = to_arrays_dict(def_func)
        self.sizes = None

    def _get_defaults(self):
        """Get the list of default values from a data list."""
        args, _, _, defaults = getargspec(
            self.py_func
        )  # pylint: disable=deprecated-method
        if defaults is None:
            return {}
        n_def = len(defaults)
        args_dict = {args[-n_def:][i]: defaults[i] for i in range(n_def)}
        return args_dict

    def _run(self):
        """Run the discipline."""
        input_vals = self.get_input_data()
        out_vals = self.py_func(**input_vals)
        if len(self.out_names) == 1:
            out_dict = {self.out_names[0]: out_vals}
        else:
            out_dict = dict(zip(self.out_names, out_vals))

        self.store_local_data(**out_dict)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the jacobian.

        :param inputs: input data.
        :type inputs: dict.
        :param outputs: output data.
        :type outputs: dict.
        """
        if self.py_jac is None:
            raise RuntimeError("Analytic jacobian is not provided !")
        if self.sizes is None:
            self.sizes = {k: v.size for k, v in self.local_data.items()}

        input_vals = self.get_input_data()
        flat_jac = self.py_jac(**input_vals)
        flat_jac = atleast_2d(flat_jac)
        self.jac = DataConversion.jac_2dmat_to_dict(
            flat_jac, self.out_names, self.in_names, self.sizes
        )

    @staticmethod
    def get_return_spec_fromstr(
        return_line,
    ):  # pylint: disable=inconsistent-return-statements
        """Get the specifications returned by a python function.

        :param return_line: the python line containing return statement
        :type return_line: str
        :return: returned string output specifications
        :rtype: str
        """
        line_cln = return_line.strip()
        if line_cln.startswith("return "):
            line_cln = line_cln.replace("return ", "")
            line_cln = re.sub(r"\s+", "", line_cln)
            outs = line_cln.split(",")
            return outs

    @staticmethod
    def _get_return_spec(func):
        """Get the specifications returned by a python function.

        :param func: the python function to be used to generate
            the MDODiscipline
        :type func: function
        :return: returned string output specifications or None
        :rtype: str or None
        """
        source = getsource(func)
        outs = None
        for line in source.split("\n"):
            outs_loc = AutoPyDiscipline.get_return_spec_fromstr(line)
            if outs_loc is not None and outs is not None:
                if tuple(outs_loc) != tuple(outs):
                    raise ValueError(
                        "Inconsistent definition of return "
                        + "statements in functions :"
                        + str(tuple(outs_loc))
                        + " != "
                        + str(tuple(outs))
                    )
            if outs_loc is not None:
                outs = outs_loc
        return outs


class AutoDiscDataProcessor(DataProcessor):
    """A data preprocessor that converts all |g| scalar input data to floats, and
    converts all discipline output data to numpy arrays."""

    def __init__(self, out_names):
        """Constructor.

        :param out_names: names of the outputs
        :type out_names: list(str)
        """
        super(AutoDiscDataProcessor, self).__init__()
        self.out_names = out_names
        self.one_output = len(out_names) == 1

    def pre_process_data(self, data):
        """Execute a pre processing of input data after they are checked by
        MDODiscipline.check_data, and before the _run method of the discipline is
        called.

        :param data: the input data to process.
        :type data: dict
        :returns: the processed input data
        :rtype: dict
        """
        processed_data = data.copy()
        for key, val in data.items():
            if len(val) == 1:
                processed_data[key] = float(val[0])
        return processed_data

    def post_process_data(self, data):
        """Execute a post processing of discipline output data after the _run method of
        the discipline, before they are checked by  MDODiscipline.check_output_data,

        :param data: the output data to process.
        :type data: dict
        :returns: the processed output data.
        :rtype: dict
        """
        processed_data = data.copy()
        for out_n, out_v in processed_data.items():
            if not isinstance(out_v, ndarray):
                out_v = array([out_v])
                processed_data[out_n] = out_v
        return processed_data


def to_arrays_dict(in_dict):
    """Ensure that a dict of data values are arrays.

    :param in_dict: the dict to be ensured.
    :returns: ensured data dict
    :rtype: dict
    """
    for out_n, out_v in in_dict.items():
        if not isinstance(out_v, ndarray):
            in_dict[out_n] = array([out_v])
    return in_dict



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#        :author: Antoine Dechaume
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Pretty string utils
===================
"""
from collections import namedtuple
from contextlib import contextmanager

from gemseo.utils.py23_compat import PY2

if PY2:
    from collections import Iterable, Mapping
else:
    from collections.abc import Iterable, Mapping

# to store the raw ingredients of a string to be formatted later
MessageLine = namedtuple("MessageLine", "str_format level args kwargs")


def pretty_repr(obj, **kwargs):
    """String representation of an object.

    :param obj: object.
    :return: pretty string representation.
    :rtype: str
    """
    delimiter = kwargs.get("delimiter", ", ")

    if isinstance(obj, Mapping):
        return delimiter.join(
            ["{}={}".format(key, repr(val)) for key, val in sorted(obj.items())]
        )

    if isinstance(obj, Iterable):
        return delimiter.join([str(val) for val in obj])

    return repr(obj)


class MultiLineString(object):
    """Multi-line string lazy evaluator.

    The creation of the string is postponed to when an instance is stringified through
    the __repr__ method. This is mainly used for logging complex strings or objects
    where the string evaluation cost may be avoided when the logging level dismisses a
    logging message.
    """

    INDENTATION = " " * 3
    DEFAULT_LEVEL = 0

    def __init__(self):
        self.__lines = []
        self.__level = None
        self.reset()

    def add(self, str_format, *args, **kwargs):
        """Add a line.

        :param str str_format: string to be process by the format() method.
        :param args: args passed to the format() method.
        :param kwargs: kwargs passed to the format() method.
        """
        self.__lines.append(MessageLine(str_format, self.__level, args, kwargs))

    def reset(self):
        """Reset the indentation."""
        self.__level = self.DEFAULT_LEVEL

    def indent(self):
        """Increase the indentation."""
        self.__level += 1

    def dedent(self):
        """Decrease the indentation."""
        if self.__level > 0:
            self.__level -= 1

    def __repr__(self):
        lines = []
        for line in self.__lines:
            str_format = self.INDENTATION * line.level + line.str_format
            lines.append(str_format.format(*line.args, **line.kwargs))
        return "\n".join(lines)

    @classmethod
    @contextmanager
    def offset(cls):
        cls.DEFAULT_LEVEL += 1
        try:
            yield
        finally:
            cls.DEFAULT_LEVEL -= 1



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Functional operations
*********************
"""
from __future__ import division, unicode_literals

from numpy import delete, insert

from gemseo.core.function import MDOFunction


class RestrictedFunction(MDOFunction):
    """Restrict an MDOFunction to a subset of its input vector Fixes the rest of the
    indices."""

    def __init__(self, orig_function, restriction_indices, restriction_values):
        """Constructor.

        :param orig_function: the original function to restrict
        :param restriction_indices: indices array of the input vector to fix
        :param restriction_values: the values of the input vector at indices
            'restriction_indices' are set to restriction_values
        """
        if not restriction_indices.shape == restriction_values.shape:
            raise ValueError(
                "Inconsistent shapes for restriction" + "values and indices "
            )
        self.restriction_values = restriction_values
        func, jac = self._restrict(orig_function, restriction_indices)
        self._orig_function = orig_function
        name = str(orig_function.name) + "_restr"
        super(RestrictedFunction, self).__init__(
            func,
            name,
            jac=jac,
            f_type=orig_function.f_type,
            expr=orig_function.expr,
            args=orig_function.args,
            dim=orig_function.dim,
            outvars=orig_function.outvars,
        )

    def _restrict(self, orig_function, restriction_indices):
        """Restricts the function, builds the pointer to f and jac.

        @param orig_function: original MDOFunction pointer
        @param restriction_indices: indices array for restriction
        """

        def restricted_function(x_vect):
            """Wrapped provided function in order to give to optimizer.

            :param x_vect: design variable
            :returns: evaluation of function at x_vect
            """
            x_full = insert(x_vect, restriction_indices, self.restriction_values)
            return orig_function(x_full)

        def restricted_jac(x_vect):
            """Wrapped provided jacobian in order to give to optimizer.

            :param x_vect: design variable
            :returns: evaluation of jacobian at x_vect
            """
            x_full = insert(x_vect, restriction_indices, self.restriction_values)
            jac = orig_function.jac(x_full)
            jac = delete(jac, restriction_indices, axis=0)
            return jac

        return restricted_function, restricted_jac


class LinerarComposition(MDOFunction):
    """Composes a function with a linear operator defined by a matrix computes
    orig_f(Mat.dot(x))"""

    def __init__(self, orig_function, interp_operator):
        """Constructor.

        :param orig_function: the original function to restrict
        :param interp_operator: operator matrix, the output of the
            function will be f(interp_operator.dot(x))
        """
        func, jac = self._restrict(orig_function, interp_operator)
        self._orig_function = orig_function
        super(LinerarComposition, self).__init__(
            func,
            str(orig_function.name) + "_comp",
            jac=jac,
            f_type=orig_function.f_type,
            expr="Mat*" + str(orig_function.expr),
            args=orig_function.args,
            dim=orig_function.dim,
            outvars=orig_function.outvars,
        )

    @staticmethod
    def _restrict(orig_function, interp_operator):
        """Generates the function restriction.

        @param orig_function : the original function to restrict
        @param interp_operator: operator matrix, the output of the
            function will be f(interp_operator.dot(x))
        """

        def restricted_function(x_vect):
            """Wrapped provided function in order to give to optimizer.

            :param x_vect: design variable
            :returns: evaluation of function at x_vect
            """
            x_full = interp_operator.dot(x_vect)
            return orig_function(x_full)

        def restricted_jac(x_vect):
            """Wrapped provided jacobian in order to give to optimizer.

            :param x_vect: design variable
            :returns: evaluation of jacobian at x_vect
            """
            x_full = interp_operator.dot(x_vect)
            jac = orig_function.jac(x_full)
            return interp_operator.T.dot(jac)

        return restricted_function, restricted_jac



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A constraints plot."""
from __future__ import division, unicode_literals

import logging
from typing import Sequence

from gemseo.post.dataset.lines import Lines
from gemseo.post.opt_post_processor import OptPostProcessor

LOGGER = logging.getLogger(__name__)


class BasicHistory(OptPostProcessor):
    """Plot the history of selected constraint, objective and observable functions.

    This post-processor requires the names of these selected outputs.
    """

    def _plot(
        self,
        data_list,  # type: Sequence[str]
    ):  # type: (...) -> None
        """
        Args:
            data_list: The names of the variables.
        """
        dataset = self.opt_problem.export_to_dataset(
            "OptimizationProblem", opt_naming=False, by_group=False
        )
        plot = Lines(dataset)
        plot.font_size = 12
        plot.xlabel = "Iterations"
        plot.figsize_x = 11
        plot.figsize_y = 6
        plot.title = "History plot"
        figures = plot.execute(save=False, show=False, variables=data_list)
        for figure in figures:
            self._add_figure(figure)



# -*- coding: utf-8 -*-
# Copyright (c) 2009-2014, Luke Maurits <luke@maurits.id.au>
# All rights reserved.
# With contributions from:
# * Chris Clark
#  * Klein Stephane
#  * John Filleau
# PTable is forked from original Google Code page in April, 2015, and now
# maintained by Kane Blueriver <kxxoling@gmail.com>.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# Taken from :
# https://github.com/kxxoling/PTable
# https://pypi.python.org/pypi/PTable
"""
Pretty table factory
********************
"""
from __future__ import ( division, 
                        unicode_literals)

import csv


from ._compact import HTMLParser
from .prettytable import PrettyTable




def from_csv(filepath, field_names=None, **kwargs):
    """Loads table from CSV file

    :param filepath: param field_names:  (Default value = None)
    :param field_names:  (Default value = None)

    """
    fmtparams = {}
    for param in ["delimiter", "doublequote", "escapechar", "lineterminator",
                  "quotechar", "quoting", "skipinitialspace", "strict"]:
        if param in kwargs:
            fmtparams[param] = kwargs.pop(param)
    if fmtparams:
        reader = csv.reader(filepath, **fmtparams)
    else:
        dialect = csv.Sniffer().sniff(filepath.read(1024))
        filepath.seek(0)
        reader = csv.reader(filepath, dialect)

    table = PrettyTable(**kwargs)
    if field_names:
        table.field_names = field_names
        next(reader)
    else:
        table.field_names = [x.strip() for x in next(reader)]

    for row in reader:
        table.add_row([x.strip() for x in row if len(x.strip()) > 0])

    return table


def from_db_cursor(cursor, **kwargs):
    """

    :param cursor:
    :param kwargs:

    """
    if cursor.description:
        table = PrettyTable(**kwargs)
        table.field_names = [col[0] for col in cursor.description]
        for row in cursor.fetchall():
            table.add_row(row)
        return table
    return None


class TableHandler(HTMLParser):
    """TableHandler"""

    def __init__(self, **kwargs):
        HTMLParser.__init__(self)
        self.kwargs = kwargs
        self.tables = []
        self.last_row = []
        self.rows = []
        self.max_row_width = 0
        self.active = None
        self.last_content = ""
        self.is_last_row_header = False
        self.colspan = 0

    def handle_starttag(self, tag, attrs):
        """

        :param tag: param attrs:
        :param attrs:

        """
        self.active = tag
        if tag == "th":
            self.is_last_row_header = True
        for (key, value) in attrs:
            if key == "colspan":
                self.colspan = int(value)

    def handle_endtag(self, tag):
        """

        :param tag:

        """
        if tag in ["th", "td"]:
            stripped_content = self.last_content.strip()
            self.last_row.append(stripped_content)
            if self.colspan:
                for _ in range(1, self.colspan):
                    self.last_row.append("")
                self.colspan = 0

        if tag == "tr":
            self.rows.append(
                (self.last_row, self.is_last_row_header))
            self.max_row_width = max(self.max_row_width, len(self.last_row))
            self.last_row = []
            self.is_last_row_header = False
        if tag == "table":
            table = self.generate_table(self.rows)
            self.tables.append(table)
            self.rows = []
        self.last_content = " "
        self.active = None

    def handle_data(self, data):
        """

        :param data:

        """
        self.last_content += data

    def generate_table(self, rows):
        """Generates from a list of rows a PrettyTable object.

        :param rows:

        """
        table = PrettyTable(**self.kwargs)
        for row in self.rows:
            if len(row[0]) < self.max_row_width:
                appends = self.max_row_width - len(row[0])
                for _ in range(1, appends):
                    row[0].append("-")

            if row[1] is True:
                self.make_fields_unique(row[0])
                table.field_names = row[0]
            else:
                table.add_row(row[0])
        return table

    @staticmethod
    def make_fields_unique(fields):
        """iterates over the row and make each field unique

        :param fields: fields
        """
        for i in range(0, len(fields)):
            for j in range(i + 1, len(fields)):
                if fields[i] == fields[j]:
                    fields[j] += "'"


def from_html(html_code, **kwargs):
    """Generates a list of PrettyTables from a string of HTML code.

    Each <table> in the HTML becomes one PrettyTable object.

    :param html_code: HTML code
    :type html_code: str
    :param kwargs: additional arguments
    """

    parser = TableHandler(**kwargs)
    parser.feed(html_code)
    return parser.tables


def from_html_one(html_code, **kwargs):
    """Generates a PrettyTables from a string of HTML code which contains only
    a single <table>

    :param html_code: HTML code
    :type html_code: str
    :param kwargs: additional arguments
    """

    tables = from_html(html_code, **kwargs)
    try:
        assert len(tables) == 1
    except AssertionError:
        raise Exception("More than one <table> in provided HTML code!"
                        + "  Use from_html instead.")
    return tables[0]


def from_md(markdown, **kwargs):
    """Generate PrettyTable from markdown string.

    :param markdown: markdown type string.
    :param kwargs: additional arguments
    :returns: a PrettyTable object.

    """
    rows = markdown.split('\n')
    title_row = rows[0]
    content_rows = rows[2:]
    table = PrettyTable(**kwargs)
    table.field_names = split_md_row(title_row)
    list(map(table.add_row, list(map(split_md_row,
                                     [x for x in content_rows if x]))))
    return table


def strip_md_content(markdown):
    """Strip the blank space and `:` in markdown table content cell.

    :param markdown: a row of markdown table
    :returns: stripped content cell
    """
    return markdown.strip().strip(':').strip()


def split_md_row(row):
    """Split markdown table.

    :param row: a row of markdown table
    :returns: Split content list
    """
    return [strip_md_content(s) for s in row.strip('|').split('|')]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A factory to instantiate MDA from their class names."""
from __future__ import division, unicode_literals

import logging

from gemseo.core.factory import Factory
from gemseo.mda.mda import MDA

LOGGER = logging.getLogger(__name__)


class MDAFactory(object):
    """MDA factory to create the MDA from a name or a class."""

    def __init__(self):
        """Initializes the factory: scans the directories to search for subclasses of
        MDA.

        Searches in "GEMSEO_PATH" and gemseo.mda
        """
        self.factory = Factory(MDA, ("gemseo.mda",))

    def create(self, mda_name, disciplines, **options):
        """Create a MDA.

        :param mda_name: name of the MDA (its classname)
        :param disciplines: list of the disciplines
        :param options: additional options specific
            to the MDA
        """

        return self.factory.create(mda_name, disciplines=disciplines, **options)

    @property
    def mdas(self):
        """Lists the available classes.

        :returns : the list of classes names
        """
        return self.factory.classes

    def is_available(self, mda_name):
        """Checks the availability of a MDA.

        :param mda_name :  name of the MDA
        :returns: True if the MDA is available
        """
        return self.factory.is_available(mda_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Binh and Korn multi-objective problem.

This module implements the Binh and Korn multi-objective problem:

.. math::

   \begin{aligned}
   \text{minimize the objective function } & f_1(x, y) = 4x^2 + 4y^2 \\
   & f_2(x, y) = (x-5)^2 + (y-5)^2 \\
   \text{with respect to the design variables }&x,\,y \\
   \text{subject to the general constraints }
   & g_1(x,y) = (x-5)^2 + y^2 \leq 25.0\\
   & g_2(x, y) = (x-8)^2 + (y+3)^2 \geq 7.7\\
   \text{subject to the bound constraints }
   & 0 \leq x \leq 5.0\\
   & 0 \leq y \leq 3.0
   \end{aligned}

"""
from __future__ import division, unicode_literals

import logging
from typing import Tuple

from numpy import array, ndarray, zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.function import MDOFunction

LOGGER = logging.getLogger(__name__)


class BinhKorn(OptimizationProblem):
    """Binh and Korn optimization problem.

    The constructor initializes the BinhKorn :class:`.OptimizationProblem`
    by defining the :class:`.DesignSpace`,
    the objective function and the constraints.

    Attributes:
        objective (MDOFunction): The objective function.
    """

    def __init__(
        self, initial_values=(1.0, 1.0)  # type: Tuple[float, float]
    ):
        """
        Args:
            initial_values: Initial value of the design variables.
        """
        design_space = DesignSpace()
        design_space.add_variable("x", 1, l_b=0.0, u_b=5.0, value=initial_values[0])
        design_space.add_variable("y", 1, l_b=0.0, u_b=3.0, value=initial_values[1])

        super(BinhKorn, self).__init__(design_space)
        self.objective = MDOFunction(
            self.__compute_binhkorn,
            name="compute_binhkorn",
            f_type="obj",
            jac=self.__compute_binhkorn_jac,
            expr="(4*x**2+ 4*y**2, (x-5.)**2 + (y-5.)**2)",
            args=["x", "y"],
        )
        ineq1 = MDOFunction(
            self.__compute_ineq_constraint1,
            name="ineq1",
            f_type="ineq",
            jac=self.__compute_ineq_constraint1_jac,
            expr="(x-5.)**2 + y**2 <= 25.",
            args=["x", "y"],
        )
        self.add_ineq_constraint(ineq1)

        ineq2 = MDOFunction(
            self.__compute_ineq_constraint2,
            name="ineq2",
            f_type="ineq",
            jac=self.__compute_ineq_constraint2_jac,
            expr="(x-8.)**2 + (y+3)**2 >= 7.7",
            args=["x", "y"],
        )
        self.add_ineq_constraint(ineq2)

    def __compute_binhkorn(
        self,
        x_dv,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the objective of analytical function.

        Args:
            x_dv: The design variable vector.

        Returns:
            The objective function value.
        """
        obj = array([0.0, 0.0])
        obj[0] = 4 * x_dv[0] ** 2 + 4 * x_dv[1] ** 2
        obj[1] = (x_dv[0] - 5.0) ** 2 + (x_dv[1] - 5.0) ** 2
        return obj

    def __compute_ineq_constraint1(
        self, x_dv  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the first constraint function.

        Args:
            x_dv: The design variable vector.

        Returns:
            The first constraint value.
        """
        return array([(x_dv[0] - 5.0) ** 2 + x_dv[1] - 25.0])

    def __compute_ineq_constraint2(
        self, x_dv  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the first constraint function.

        Args:
            x_dv: The design variable vector.

        Returns:
            The second constraint value.
        """
        return array([-((x_dv[0] - 8.0) ** 2) - (x_dv[1] + 3) + 7.7])

    @staticmethod
    def __compute_binhkorn_jac(
        x_dv,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the gradient of objective.

        Args:
            x_dv: The design variables vector.

        Returns:
            The gradient of the objective functions
            w.r.t the design variables
        """
        jac = zeros([2, 2])
        jac[0, 0] = 8.0 * x_dv[0]
        jac[0, 1] = 8.0 * x_dv[1]
        jac[1, 0] = 2.0 * x_dv[0] - 10.0
        jac[1, 1] = 2.0 * x_dv[1] - 10.0
        return jac

    @staticmethod
    def __compute_ineq_constraint1_jac(
        x_dv,  # type: ndarray
    ):  # (...) -> ndarray
        """Compute the first inequality constraint jacobian.

        Args:
            x_dv: The design variables vector.

        Returns:
            The gradient of the first constraint function
            w.r.t the design variables.
        """
        jac = zeros([1, 2])
        jac[0, 0] = 2.0 * x_dv[0] - 10.0
        jac[0, 1] = 2.0 * x_dv[1]
        return jac

    @staticmethod
    def __compute_ineq_constraint2_jac(
        x_dv,  # type: ndarray
    ):  # (...) -> ndarray
        """Compute the second inequality constraint jacobian.

        Args:
            x_dv: The design variables vector.

        Returns:
            The gradient of the second constraint function
            w.r.t the design variables.
        """
        jac = zeros([1, 2])
        jac[0, 0] = -2.0 * x_dv[0] + 16.0
        jac[0, 1] = -2.0 * x_dv[1] + 6.0
        return jac



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import division, unicode_literals

from copy import deepcopy
from functools import reduce as f_reduce

from numpy import array, hstack, ndarray, vstack, zeros

"""
Convert arrays/dict of arrays data
**********************************
"""


class DataConversion(object):
    """This class contains static methods that convert:

    - data dict into numpy array
    - numpy array into data dict by updating an existing dict
    """

    FLAT_JAC_SEP = "!d$_$d!"

    @staticmethod
    def dict_to_array(data_dict, data_names):
        """Convert a dict into an array.

        :param data_dict: the data dictionary
        :param data_names: the data keys to concatenate
        :returns: an array
        :rtype: ndarray
        """
        if not data_names:
            return array([])
        values_list = [data_dict[name] for name in data_names]
        return hstack(values_list)

    @staticmethod
    def list_of_dict_to_array(data_list, data_names, group=None):
        """Convert a list of dict into an array.

        :param data_list: the data list
        :param data_names: the data keys to concatenate
        :param group: groupe of keys to concatenate
        :returns: an array
        :rtype: ndarray
        """
        dict_to_array = DataConversion.dict_to_array
        if group is not None:
            out_array = vstack(
                [dict_to_array(element[group], data_names) for element in data_list]
            )
        else:
            out_array = array(
                [dict_to_array(element, data_names) for element in data_list]
            )
        return out_array

    @staticmethod
    def array_to_dict(data_array, data_names, data_sizes):
        """Convert an array into a dict.

        :param data_array: the array
        :param data_names: list of names (keys of the resulting dict)
        :param data_sizes: dict of (name, size)
        :returns: a dict
        :rtype: dict
        """
        current_position = 0
        array_dict = {}
        if len(data_array.shape) == 2:
            for name in data_names:
                array_dict[name] = data_array[
                    :, current_position : current_position + data_sizes[name]
                ]
                current_position += data_sizes[name]
        elif len(data_array.shape) == 1:
            for name in data_names:
                array_dict[name] = data_array[
                    current_position : current_position + data_sizes[name]
                ]
                current_position += data_sizes[name]
        else:
            raise ValueError("Invalid data dimension >2 !")
        return array_dict

    @staticmethod
    def jac_2dmat_to_dict(flat_jac, outputs, inputs, data_sizes):
        """Converts a full 2D jacobian to a dict of dict sparse format.

        :param flat_jac : the flat 2D jacobian
        :param inputs: derive outputs wrt inputs
        :param outputs: outputs to be derived
        :param data_sizes: dict of (name, size) for names in inputs and outputs
        """
        curr_out = 0
        jac_dict = {}
        for out in outputs:
            jac_dict[out] = {}
            out_size = data_sizes[out]
            curr_in = 0
            for inpt in inputs:
                inpt_size = data_sizes[inpt]
                jac_dict[out][inpt] = flat_jac[
                    curr_out : curr_out + out_size, curr_in : curr_in + inpt_size
                ]
                curr_in += inpt_size
            curr_out += out_size
        return jac_dict

    @staticmethod
    def jac_3dmat_to_dict(jac, outputs, inputs, data_sizes):
        """Converts a 3D jacobian (list of 2D jacobians) to a dict of dict sparse
        format.

        :param jac : list of 2D jacobians
        :param inputs: derive outputs wrt inputs
        :param outputs: outputs to be derived
        :param data_sizes: dict of (name, size) for names in inputs and outputs
        """
        curr_out = 0
        jac_dict = {}
        for out in outputs:
            jac_dict[out] = {}
            out_size = data_sizes[out]
            curr_in = 0
            for inpt in inputs:
                inpt_size = data_sizes[inpt]
                jac_dict[out][inpt] = jac[
                    :, curr_out : curr_out + out_size, curr_in : curr_in + inpt_size
                ]
                curr_in += inpt_size
            curr_out += out_size
        return jac_dict

    @staticmethod
    def dict_jac_to_2dmat(jac_dict, outputs, inputs, data_sizes):
        """Converts a dict of dict sparse format jacobian to a full 2D jacobian.

        :param jac_dict : the jacobian dict of dict
        :param inputs: derive outputs wrt inputs
        :param outputs: outputs to be derived
        :param data_sizes: dict of (name, size) for names in inputs and outputs
        """
        n_outs = sum((data_sizes[out] for out in outputs))
        n_inpts = sum((data_sizes[inpt] for inpt in inputs))
        flat_jac = zeros((n_outs, n_inpts))
        curr_out = 0
        for out in outputs:
            out_size = data_sizes[out]
            curr_in = 0
            for inpt in inputs:
                inpt_size = data_sizes[inpt]
                flat_jac[
                    curr_out : curr_out + out_size, curr_in : curr_in + inpt_size
                ] = jac_dict[out][inpt]
                curr_in += inpt_size
            curr_out += out_size
        return flat_jac

    @staticmethod
    def dict_jac_to_dict(jac_dict):
        """Converts a full 2D jacobian to a flat dict.

        :param jac_dict : the jacobian dict of dict
        """

        flat_jac = {}
        for out_name, jac_dict_loc in jac_dict.items():
            for inpt_name, jac_mat in jac_dict_loc.items():
                flat_name = DataConversion.flat_jac_name(out_name, inpt_name)
                flat_jac[flat_name] = jac_mat
        return flat_jac

    @staticmethod
    def flat_jac_name(out_name, inpt_name):
        """get the flat jacobian name from the full jacobian name.

        :param out_name: name of output
        :param inpt_name: name of input
        """
        return out_name + DataConversion.FLAT_JAC_SEP + inpt_name

    @staticmethod
    def dict_to_jac_dict(flat_jac_dict):
        """Converts a flat dict to full 2D jacobian.

        :param flat_jac_dict : the jacobian dict
        """

        jac = {}
        sep = DataConversion.FLAT_JAC_SEP
        all_outs = set((key.split(sep)[0]) for key in flat_jac_dict)
        all_ins = set((key.split(sep)[1]) for key in flat_jac_dict)

        for out_name in all_outs:
            jac[out_name] = {}
            for inpt_name in all_ins:
                flat_name = DataConversion.flat_jac_name(out_name, inpt_name)
                jac[out_name][inpt_name] = flat_jac_dict[flat_name]
        return jac

    @staticmethod
    def update_dict_from_array(reference_input_data, data_names, values_array):
        """Updates a data dictionary from values array The order of the data in the
        array follows the order of the data names.

        :param reference_input_data: the base input data dict
        :param data_names: the dict keys to be updated
        :param values_array: the data array to update the dictionary
        :returns: the updated data dict
        """
        if not isinstance(values_array, ndarray):
            raise TypeError(
                "Values array must be a numpy.ndarray, "
                + "got instead :"
                + str(type(values_array))
            )
        data = deepcopy(reference_input_data)
        if not data_names:
            return data
        i_min = 0
        for key in data_names:
            value_ref = reference_input_data.get(key)
            if value_ref is None:
                raise ValueError("Reference data has no item named: " + str(key))
            value_ref = array(reference_input_data[key], copy=False)
            shape_ref = value_ref.shape
            value_flatten = value_ref.flatten()  # copy is made here
            i_max = i_min + len(value_flatten)
            if len(values_array) < i_max:
                raise ValueError(
                    "Inconsistent input array size  of values array "
                    + str(values_array)
                    + " with reference data shape "
                    + str(shape_ref)
                    + "for data named:"
                    + str(key)
                )
            value_flatten[:] = values_array[i_min:i_max]
            data[key] = value_flatten.reshape(shape_ref)
            i_min = i_max
        if i_max != values_array.size:
            raise ValueError(
                "Inconsistent data shapes !\n"
                + "Could not use the whole data array of shape "
                + str(values_array.shape)
                + " (only reached max index ="
                + str(i_max)
                + "),\nwhile updating data dictionary keys "
                + str(data_names)
                + "\n of shapes : "
                + str([(k, reference_input_data[k].shape) for k in data_names])
            )
        return data

    @staticmethod
    def deepcopy_datadict(data_dict, keys=None):
        """Performs a deepcopy of a data dict treats numpy arrays specially using
        array.copy() instead of deepcopy.

        :param data_dict: data dict to copy
        """
        dict_cp = {}
        if keys is None:
            for key, val in data_dict.items():
                if isinstance(val, ndarray):
                    dict_cp[key] = val.copy()
                else:
                    dict_cp[key] = deepcopy(val)
        else:
            common_keys = set(keys) & set(data_dict.keys())
            for key in common_keys:
                val = data_dict[key]
                if isinstance(val, ndarray):
                    dict_cp[key] = val.copy()
                else:
                    dict_cp[key] = deepcopy(val)
        return dict_cp

    @staticmethod
    def __set_reduce(s_1, s_2):
        """Returns a set of unique elements of merged s_1 and s_2."""
        set(s_1)
        set(s_2)
        return set(s_1) | set(s_2)

    @staticmethod
    def get_all_inputs(disciplines, recursive=False):
        """Lists all the inputs of the disciplines Merges the input data from the
        disicplines grammars.

        :param disciplines: the list of disciplines to search
        :param recursive: if True, searches for the inputs of the
            sub disciplines (when some disciplines are scenarios)
        :returns: the list of input data
        """

        sub_d = [disc for disc in disciplines if not disc.is_scenario()]
        if recursive:
            sub_sc = [disc for disc in disciplines if disc.is_scenario()]

            # Take all sub disciplines
            flat_sub_d = list(
                f_reduce(
                    DataConversion.__set_reduce, (scen.disciplines for scen in sub_sc)
                )
            )
            # Take disciplines plus sub disciplines
            flat_sub_d = flat_sub_d + sub_d
            return DataConversion.get_all_inputs(flat_sub_d, recursive=False)

        data = f_reduce(
            DataConversion.__set_reduce, (disc.get_input_data_names() for disc in sub_d)
        )
        return list(data)

    @staticmethod
    def get_all_outputs(disciplines, recursive=False):
        """Lists all the outputs of the disciplines Merges the output data from the
        disciplines grammars.

        :param disciplines: the list of disciplines to search
        :param recursive: if True, searches for the outputs of the
            sub disciplines (when some disciplines are scenarios)

        :returns: the list of output data
        """

        sub_d = [disc for disc in disciplines if not disc.is_scenario()]
        if recursive:
            sub_sc = [disc for disc in disciplines if disc.is_scenario()]

            # Take all sub disciplines
            flat_sub_d = list(
                f_reduce(
                    DataConversion.__set_reduce, (scen.disciplines for scen in sub_sc)
                )
            )
            # Take disciplines plus sub disciplines
            flat_sub_d = flat_sub_d + sub_d
            return DataConversion.get_all_outputs(flat_sub_d, recursive=False)

        data = f_reduce(
            DataConversion.__set_reduce,
            (disc.get_output_data_names() for disc in sub_d),
        )
        return list(data)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# -*-mode: python; py-indent-offset: 4; tab-width: 8; coding:utf-8 -*-
# Copyright (c) 2018 IRT-AESE.
# All rights reserved.
#
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Arthur Piat
#        :author: François Gallard: initial author of the scilab version
#                                   of MatlabDataProcessorWrapper
#        :author: Nicolas Roussouly: GEMSEO integration

#
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Definition of Matlab data processor.

Overview
--------

The class and functions in this module enables to
manipulate data from and toward the Matlab workspace.
It also enables to read and write Matlab data file (.mat).
"""

from typing import Mapping, Union

import matlab
import scipy.io
from numpy import array, iscomplexobj, ndarray

from gemseo.core.data_processor import DataProcessor
from gemseo.utils.py23_compat import Path


class MatlabDataProcessor(DataProcessor):
    """A Matlab data processor.

    Convert GEMSEO format to Matlab format.

    Examples:
        >>> # Build a new instance
        >>> proc = MatlabDataProcessor()
        >>> # initial python data
        >>> d = {"x": array([2]), "y": array([2j], dtype="complex")}
        >>> # process data to matlab format
        >>> res = proc.pre_process_data(d)
        >>> print(res)
        >>>
        >>> # initial data in matlab format
        >>> d = {"y": double([2, 3]), "x": double([2j], is_complex=True)}
        >>> # process to Python format
        >>> res = proc.post_process_data(d)
        >>> print(res)
    """

    def pre_process_data(
        self, data  # type: Mapping[str, ndarray]
    ):  # type: (...) -> Mapping[str, matlab.double]
        """Transform data from GEMSEO to Matlab.

        The function takes a dict of ndarray and return
        a dict of matlab.double.

        Args:
            data: The input data.

        Returns:
            The data with matlab array types.
        """
        processed_data = {}

        for name, values in data.items():
            if isinstance(values, matlab.double):
                processed_data[name] = values
            else:
                processed_data[name] = array2double(values)

        return processed_data

    def post_process_data(
        self, data  # type: Mapping[str, matlab.double]
    ):  # type: (...) -> Mapping[str, ndarray]
        """Transform the output data from Matlab to GEMSEO.

        Args:
            data: The data with matlab arrays.

        Returns:
            The data with numpy arrays.
        """
        processed_data = {}

        for name, values in data.items():
            if isinstance(values, ndarray):
                processed_data[name] = values.copy()
            else:
                processed_data[name] = double2array(values)

        return processed_data


def load_matlab_file(
    file_path,  # type: Union[str, Path]
):  # type: (...) -> Mapping[str, matlab.double]
    """Read .mat file and convert it to usable format for Matlab.

    Args:
        file_path: The path to a .mat file.

    Returns:
        The dict of matlab.double.
    """
    row_data = scipy.io.loadmat(str(file_path))
    clean_data = {}
    for parameter in row_data:
        if parameter not in ["__header__", "__globals__", "__version__"]:
            clean_data[parameter] = array2double(row_data[parameter])

    return clean_data


def save_matlab_file(
    dict_to_save,  # type: dict
    file_path="output_dict",  # type: Union[str, Path]
    *args,  # type: bool
    **kwargs  # type: bool
):  # type: (...) -> None
    """Create a .mat file from dict of ndarray.

    Args:
        dict_to_save: The dict of ndarray to be saved.
        file_path: The path where to sabe the file.
        *args: The list of scipy.io.savemat options.
        **kwargs: The dict of scipy.io.savemat options.

    Raises:
        ValueError: If the saved dictionary is nor composed
            of ndarray only.
    """
    saved_dict = dict_to_save.copy()

    for key, value in saved_dict.items():
        if isinstance(value, matlab.double):
            saved_dict[key] = double2array(value)
        elif not isinstance(value, ndarray):
            msg = "The saved dict must be composed of ndarray only"
            raise ValueError(msg)

    scipy.io.savemat(file_name=str(file_path), mdict=saved_dict, *args, **kwargs)


def array2double(
    data_array,  # type: ndarray
):  # type (..) -> matlab.double
    """Turn a ndarray into a matlab.double.

    Args:
        data_array: The numpy array to be converted.

    Returns:
        The matlab.double value.
    """
    is_cmplx = iscomplexobj(data_array)
    if len(data_array.shape) == 1:
        return matlab.double(data_array.tolist(), is_complex=is_cmplx)[0]
    else:
        return matlab.double(data_array.tolist(), is_complex=is_cmplx)


def double2array(
    matlab_double,  # type: matlab.double
):  # type: (...) -> ndarray
    """Turn a matlab double into ndarray.

    Args:
        matlab_double: The matlab.double values.

    Returns:
        The array of values.
    """
    if iscomplexobj(matlab_double):
        d_type = "complex"
    else:
        d_type = None

    # note here that we can treat string as well
    # -> we put string into an array as float
    #    (otherwise the array has no shape)
    if isinstance(matlab_double, float) or isinstance(matlab_double, str):
        output = array([matlab_double], dtype=d_type)
    else:
        output = array(matlab_double, dtype=d_type)

    if output.shape[0] == 1 and len(output.shape) > 1:
        output = output[0]

    return output


def convert_array_from_matlab(
    data,  # type: Mapping[str, matlab.double]
):  # type: (...) -> Mapping[str, ndarray]
    """Convert dict of matlab.output to dict of ndarray.

    Args:
        data: The dict of matlab.double.

    Returns:
        The dict of ndarray.
    """
    output_values = {}
    for matlab_key in data:
        current_value = data[matlab_key]
        output_values[matlab_key] = double2array(current_value)
    return output_values


def convert_array_to_matlab(
    data,  # type: Mapping[str, ndarray]
):  # type: (...) -> Mapping[str, matlab.double]
    """Convert gems dict of ndarray to dict of matlab.double.

    Args:
        data: The dict of ndarray.

    Returns:
        The dict of matlab.double.
    """
    output = {}
    for keys in data:
        current_data = data[keys]
        if not (len(current_data) == 1):
            output[keys] = array2double(current_data)
        else:
            if iscomplexobj(current_data):
                output[keys] = complex(current_data)
            else:
                output[keys] = float(current_data)

    return output



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Module containing a factory to create an instance of :class:`.Distribution`."""

from __future__ import division, unicode_literals

from typing import List, Optional, Union

from gemseo.core.factory import Factory
from gemseo.uncertainty.distributions.distribution import (
    Distribution,
    ParametersType,
    StandardParametersType,
)

DistributionParametersType = Union[
    int, ParametersType, Optional[StandardParametersType], float
]


class DistributionFactory(object):
    """Factory to build instances of :class:`.Distribution`.

    At initialization, this factory scans the following modules
    to search for subclasses of this class:

    - the modules located in "gemseo.uncertainty.distributions" and its sub-packages,
    - the modules referenced in the "GEMSEO_PATH",
    - the modules referenced in the "PYTHONPATH" and starting with "gemseo_".

    Then, it can check if a class is present or return the list of available classes.

    Lastly, it can create an instance of a class.

    Examples:
        >>> from gemseo.uncertainty.distributions.factory import DistributionFactory
        >>> factory = DistributionFactory()
        >>> factory.is_available("OTNormalDistribution")
        True
        >>> factory.available_distributions[-3:]
        ['SPNormalDistribution', 'SPTriangularDistribution', 'SPUniformDistribution']
        >>> distribution = factory.create("OTNormalDistribution", "x")
        >>> print(distribution)
        Normal(mu=0.0, sigma=1.0)
    """

    def __init__(self):  # noqa: D107
        # type: (...) -> None
        self.factory = Factory(Distribution, ("gemseo.uncertainty.distributions",))

    def create(
        self,
        distribution_name,  # type: str
        variable,  # type: str
        **parameters  # type: DistributionParametersType
    ):
        # type: (...) -> Distribution
        """Create a probability distribution for a given random variable.

        Args:
            distribution_name: The name of a class defining a distribution.
            variable: The name of the random variable.
            **parameters: The parameters of the distribution.

        Returns:
            The probability distribution instance.
        """
        return self.factory.create(distribution_name, variable=variable, **parameters)

    @property
    def available_distributions(self):
        # type: (...) -> List[str]
        """The available probability distributions."""
        return self.factory.classes

    def is_available(
        self,
        distribution_name,  # type: str
    ):
        # type: (...) -> bool
        """Check the availability of a probability distribution.

        Args:
            distribution_name: The name of a class defining a distribution.

        Returns:
            The availability of the distribution.
        """
        return self.factory.is_available(distribution_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#         documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalability study - Result
==========================
"""
from __future__ import division, unicode_literals

import logging
import pickle

from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)

RESULTS_DIRECTORY = Path("results")


class ScalabilityResult(object):
    """Scalability Result."""

    def __init__(self, name, id_scaling, id_sample):
        """Constructor.

        :param str name: name of the scalability result.
        :param int id_scaling: scaling identifiant
        :param int id_sample: sample identifiant
        """
        self.name = name
        self.id_scaling = id_scaling
        self.id_sample = id_sample
        self.algo = None
        self.algo_options = None
        self.formulation_options = None
        self.formulation = None
        self.scaling = None
        self.n_calls = None
        self.n_calls_linearize = None
        self.n_calls_top_level = None
        self.n_calls_linearize_top_level = None
        self.exec_time = None
        self.original_exec_time = None
        self.status = None
        self.is_feasible = None
        self.disc_names = None
        self.old_varsizes = None
        self.new_varsizes = None
        self.output_names = None

    def get(
        self,
        algo,
        algo_options,
        formulation,
        formulation_options,
        scaling,
        n_calls,
        n_calls_linearize,
        n_calls_top_level,
        n_calls_linearize_top_level,
        exec_time,
        status,
        is_feasible,
        disc_names,
        output_names,
        old_varsizes,
        new_varsizes,
    ):
        """Get a scalability result for a given optimization strategy and a given
        scaling strategy.

        :param str algo: name of the optimization algorithm
        :param dict algo_options: options of the optimization algorithm
        :param str formulation: name of the MDO formulation
        :param dict formulation_options: options of the MDO formulation
        :param scaling: scaling strategy
        :param list(int) n_calls: number of calls for each discipline
        :param list(int) n_calls_linearize: number of linearization
            for each discipline
        :param list(int) n_calls_top_level: number of calls for each discipline
        :param list(int) n_calls_linearize_top_level: number of linearization
            for each discipline
        :param float exec_time: execution time
        :param int status: status of the optimization scenario
        :param bool is_feasible: feasibility of the optimization solution
        :param list(str) disc_names: list of discipline names
        :param dict output_names: list of output names
        :param dict old_varsizes: old variable sizes
        :param dict new_varsizes: new variable sizes
        """
        self.algo = algo
        self.algo_options = algo_options
        self.formulation = formulation
        self.formulation_options = formulation_options
        self.scaling = scaling
        self.n_calls = n_calls
        self.n_calls_linearize = n_calls_linearize
        self.n_calls_top_level = n_calls_top_level
        self.n_calls_linearize_top_level = n_calls_linearize_top_level
        self.exec_time = exec_time
        self.status = status
        self.is_feasible = is_feasible
        self.disc_names = disc_names
        self.output_names = output_names
        self.old_varsizes = old_varsizes
        self.new_varsizes = new_varsizes

    def get_file_path(self, study_directory):
        """Get file path.

        :param str study_directory: study directory name.
        """
        fname = Path(self.name).with_suffix(".pkl")
        fpath = Path(study_directory) / RESULTS_DIRECTORY / fname
        return fpath

    def save(self, study_directory):
        """Save a scalability result into a pickle file whose name is the name of the
        ScalabilityResult instance.

        :param str study_directory: study directory name.
        """
        fpath = self.get_file_path(study_directory)
        result = {
            "algo": self.algo,
            "algo_options": self.algo_options,
            "formulation": self.formulation,
            "formulation_options": self.formulation_options,
            "scaling": self.scaling,
            "n_calls": self.n_calls,
            "n_calls_linearize": self.n_calls_linearize,
            "n_calls_top_level": self.n_calls_top_level,
            "n_calls_linearize_top_level": self.n_calls_linearize_top_level,
            "exec_time": self.exec_time,
            "status": self.status,
            "is_feasible": self.is_feasible,
            "disc_names": self.disc_names,
            "output_names": self.output_names,
            "old_varsizes": self.old_varsizes,
            "new_varsizes": self.new_varsizes,
        }
        with fpath.open("wb") as fout:
            pickle.dump(result, fout)
        return fpath

    def load(self, study_directory):
        """Load a scalability result from a pickle file whose name is the name of the
        ScalabilityResult instance."""
        fname = self.name + ".pkl"
        fpath = Path(study_directory) / RESULTS_DIRECTORY / fname
        with fpath.open("rb") as fin:
            result = pickle.load(fin)
        self.algo = result["algo"]
        self.formulation = result["formulation"]
        self.scaling = result["scaling"]
        self.n_calls = result["n_calls"]
        self.n_calls_linearize = result["n_calls_linearize"]
        self.n_calls_top_level = result["n_calls_top_level"]
        self.n_calls_linearize_top_level = result["n_calls_linearize_top_level"]
        self.exec_time = result["exec_time"]
        self.status = result["status"]
        self.is_feasible = result["is_feasible"]
        self.disc_names = result["disc_names"]
        self.output_names = result["output_names"]
        self.old_varsizes = result["old_varsizes"]
        self.new_varsizes = result["new_varsizes"]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to create a triangular distribution from the SciPy library.

This class inherits from :class:`.SPDistribution`.
"""

from __future__ import division, unicode_literals

from gemseo.uncertainty.distributions.scipy.distribution import SPDistribution


class SPTriangularDistribution(SPDistribution):
    """Create a triangular distribution.

    Example:
        >>> from gemseo.uncertainty.distributions.scipy.triangular import (
        ...     SPTriangularDistribution
        ... )
        >>> distribution = SPTriangularDistribution('x', -1, 0, 1)
        >>> print(distribution)
        triang(lower=-1, mode=0, upper=1)
    """

    def __init__(
        self,
        variable,  # type: str
        minimum=0.0,  # type: float
        mode=0.5,  # type: float
        maximum=1.0,  # type: float
        dimension=1,  # type: int
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the triangular random variable.
            minimum: The minimum of the triangular random variable.
            mode: The mode of the triangular random variable.
            maximum: The maximum of the triangular random variable.
            dimension: The dimension of the triangular random variable.
        """
        parameters = {
            "loc": minimum,
            "scale": maximum - minimum,
            "c": (mode - minimum) / float(maximum - minimum),
        }
        standard_parameters = {
            self._LOWER: minimum,
            self._MODE: mode,
            self._UPPER: maximum,
        }
        super(SPTriangularDistribution, self).__init__(
            variable, "triang", parameters, dimension, standard_parameters
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Factory to create |g| disciplines
**************************************
"""
from __future__ import division, unicode_literals

from copy import deepcopy
from os.path import dirname, join

from gemseo.core import discipline
from gemseo.core.discipline import MDODiscipline
from gemseo.core.factory import Factory
from gemseo.core.json_grammar import JSONGrammar


class DisciplinesFactory(object):
    """The **DisciplinesFactory** is used to create :class:`.MDODiscipline` objects that
    are known to |g|

    Three types of directories are scanned
    to import the :class:`.MDODiscipline` classes:

    - the environment variable "GEMSEO_PATH" may contain
      the list of directories to scan,
    - the present directory (:doc:`gemseo.problems`) contains some
      benchmark test cases,
    """

    def __init__(self):
        """The constructor initializes the factory by scanning the directories to search
        for subclasses of :class:`.MDODiscipline` objects.

        Searches in "GEMSEO_PATH" and :doc:`gemseo.problems`.
        """
        # Defines the benchmark problems to be imported
        internal_modules_paths = ("gemseo.problems", "gemseo.core", "gemseo.wrappers")
        self.factory = Factory(MDODiscipline, internal_modules_paths)

        base_gram_path = join(
            dirname(discipline.__file__), "MDODiscipline_options.json"
        )
        self.__base_grammar = JSONGrammar(
            "MDODiscipline_options", schema_file=base_gram_path
        )
        self.__base_grammar_names = self.__base_grammar.get_data_names()

    def create(self, discipline_name, **options):
        """Create a :class:`.MDODiscipline` from its name.

        :param discipline_name: name of the discipline
        :type discipline_name: str
        :param options: options of the discipline,
            both the options to be passed to the constructor
            and the options that are generic to all the disciplines
        :type options: dict
        :returns: the discipline instance
        """
        com_opts_dict, spec_opts_dict = self.__filter_common_options(options)
        self.__base_grammar.load_data(com_opts_dict)
        disc = self.factory.create(discipline_name, **spec_opts_dict)
        if "linearization_mode" in com_opts_dict:
            disc.linearization_mode = com_opts_dict["linearization_mode"]

        cache_opts = self.__filter_opts_dict(com_opts_dict, "cache")
        if cache_opts:
            disc.set_cache_policy(**cache_opts)

        jac_opts = self.__filter_opts_dict(com_opts_dict, "jac")
        if jac_opts:
            disc.set_jacobian_approximation(**jac_opts)

        return disc

    @staticmethod
    def __filter_opts_dict(options, startstring):
        """Filters the options that start with a string.

        :param options: discipline options
        :type options: dict
        :param startstring: predicate for the option
        :type startstring: str
        """
        return {k: v for k, v in options.items() if k.startswith(startstring)}

    def __filter_common_options(self, options):
        """Separates options:

        - from the :class:`.MDODiscipline` options grammar
        - from the options that are specific to the discipline.

        :param options: options of the discipline
        :type options: dict
        """
        com_opts_names = self.__base_grammar_names
        com_opts_dict = {k: v for k, v in options.items() if k in com_opts_names}
        spec_opts_dict = {k: v for k, v in options.items() if k not in com_opts_dict}
        return com_opts_dict, spec_opts_dict

    def update(self):
        """Updates the paths, to be used if GEMSEO_PATH was changed."""
        self.factory.update()

    @property
    def disciplines(self):
        """Lists the available :class:`.MDODiscipline`, known to this factory.

        :returns: the list of available disciplines names
            (ie their class names)
        """
        return self.factory.classes

    def get_options_grammar(self, name, write_schema=False, schema_file=None):
        """Get the options default values for the given class name Only addresses kwargs
        Generates.

        :param name: name of the class
        :type name: str
        :param schema_file: the output json file path. If None: input.json or
            output.json depending on grammar type.
            (Default value = None)
        :type schema_file: str
        :param write_schema: if True, writes the schema files
            (Default value = False)
        :type write_schema: bool
        :returns: the json grammar for options
        """
        disc_gram = self.factory.get_options_grammar(name, write_schema, schema_file)
        base_grammar = deepcopy(self.__base_grammar)
        base_grammar.update_from(disc_gram)
        return base_grammar



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#         documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalability study - Process
===========================

The :class:`.ScalabilityStudy` class implements
the concept of scalability study:

1. By instantiating a :class:`.ScalabilityStudy`, the user defines
   the MDO problem in terms of design parameters, objective function and
   constraints.
2. For each discipline, the user adds a dataset stored
   in a :class:`.Dataset` and select a type of
   :class:`.ScalableModel` to build the :class:`.ScalableDiscipline`
   associated with this discipline.
3. The user adds different optimization strategies, defined in terms
   of both optimization algorithms and MDO formulation.
4. The user adds different scaling strategies, in terms of sizes of
   design parameters, coupling variables and equality and inequality
   constraints. The user can also define a scaling strategies according to
   particular parameters rather than groups of parameters.
5. Lastly, the user executes the :class:`.ScalabilityStudy` and the results
   are written in several files and stored into directories
   in a hierarchical way, where names depends on both MDO formulation,
   scaling strategy and replications when it is necessary. Different kinds
   of files are stored: optimization graphs, dependency matrix plots and
   of course, scalability results by means of a dedicated class:
   :class:`.ScalabilityResult`.
"""
from __future__ import division, unicode_literals

import logging
import numbers
from copy import deepcopy

from numpy import inf

from gemseo.problems.scalable.data_driven.problem import ScalableProblem
from gemseo.problems.scalable.data_driven.study.result import ScalabilityResult
from gemseo.utils.py23_compat import Path, string_types
from gemseo.utils.string_tools import MultiLineString

LOGGER = logging.getLogger(__name__)

RESULTS_DIRECTORY = Path("results")
POST_DIRECTORY = Path("visualization")
POSTOPTIM_DIRECTORY = POST_DIRECTORY / "optimization_history"
POSTSTUDY_DIRECTORY = POST_DIRECTORY / "scalability_study"
POSTSCAL_DIRECTORY = POST_DIRECTORY / "dependency_matrix"


class ScalabilityStudy(object):
    """Scalability Study."""

    def __init__(
        self,
        objective,
        design_variables,
        directory="study",
        prefix="",
        eq_constraints=None,
        ineq_constraints=None,
        maximize_objective=False,
        fill_factor=0.7,
        active_probability=0.1,
        feasibility_level=0.8,
        start_at_equilibrium=True,
        early_stopping=True,
        coupling_variables=None,
    ):
        """
        The constructor of the ScalabilityStudy class requires two mandatory
        arguments:

        - the :code:`'objective'` name,
        - the list of :code:`'design_variables'` names.

        Concerning output files, we can specify:

        - the :code:`directory` which is :code:`'study'` by default,
        - the prefix of output file names (default: no prefix).

        Regarding optimization parametrization, we can specify:

        - the list of equality constraints names (:code:`eq_constraints`),
        - the list of inequality constraints names (:code:`ineq_constraints`),
        - the choice of maximizing the objective function
          (:code:`maximize_objective`).

        By default, the objective function is minimized and the MDO problem
        is unconstrained.

        Last but not least, with regard to the scalability methodology,
        we can overwrite:

        - the default fill factor of the input-output dependency matrix
          :code:`ineq_constraints`,
        - the probability to set the inequality constraints as active at
          initial step of the optimization :code:`active_probability`,
        - the offset of satisfaction for inequality constraints
          :code:`feasibility_level`,
        - the use of a preliminary MDA to start at equilibrium
          :code:`start_at_equilibrium`,
        - the post-processing of the optimization database to get results
          earlier than final step :code:`early_stopping`.

        :param str objective: name of the objective
        :param list(str) design_variables: names of the design variables
        :param str directory: working directory of the study. Default: 'study'.
        :param str prefix: prefix for the output filenames. Default: ''.
        :param list(str) eq_constraints: names of the equality constraints.
            Default: None.
        :param list(str) ineq_constraints: names of the inequality constraints
            Default: None.
        :param bool maximize_objective: maximizing objective. Default: False.
        :param float fill_factor: default fill factor of the input-output
            dependency matrix. Default: 0.7.
        :param float active_probability: probability to set the inequality
            constraints as active at initial step of the optimization.
            Default: 0.1
        :param float feasibility_level: offset of satisfaction for inequality
            constraints. Default: 0.8.
        :param bool start_at_equilibrium: start at equilibrium
            using a preliminary MDA. Default: True.
        :param bool early_stopping: post-process the optimization database
            to get results earlier than final step.
        """
        LOGGER.info("Initialize the scalability study")
        self.prefix = prefix
        self.directory = Path(directory)
        self.__create_directories()
        self.datasets = []
        self.objective = objective
        self.design_variables = design_variables
        self.eq_constraints = eq_constraints or []
        self.ineq_constraints = ineq_constraints or []
        self.coupling_variables = coupling_variables or []
        self.maximize_objective = maximize_objective
        self.formulations = []
        self.formulations_options = []
        self.algorithms = []
        self.algorithms_options = []
        self.scalings = []
        self.var_scalings = []
        self.__check_fill_factor(fill_factor)
        self._default_fill_factor = fill_factor
        self.__check_proportion(active_probability)
        self.active_probability = active_probability
        if isinstance(feasibility_level, dict):
            for _, value in feasibility_level.items():
                self.__check_proportion(value)
        else:
            self.__check_proportion(feasibility_level)
        self.feasibility_level = feasibility_level
        self.start_at_equilibrium = start_at_equilibrium
        self.results = []
        self.early_stopping = early_stopping
        self._group_dep = {}
        self._fill_factor = {}
        self.top_level_diff = []
        self._all_data = None
        optimize = "maximize" if self.maximize_objective else "minimize"
        msg = MultiLineString()
        msg.indent()
        msg.add("Optimization problem")
        msg.indent()
        msg.add("Objective: {} {}", optimize, self.objective)
        msg.add("Design variables: {}", self.design_variables)
        msg.add("Equality constraints: {}", self.eq_constraints)
        msg.add("Inequality constraints: {}", self.ineq_constraints)
        msg.dedent()
        msg.add("Study properties")
        msg.indent()
        msg.add("Default fill factor: {}", self._default_fill_factor)
        msg.add("Active probability: {}", self.active_probability)
        msg.add("Feasibility level: {}", self.feasibility_level)
        msg.add("Start at equilibrium: {}", self.start_at_equilibrium)
        msg.add("Early stopping: {}", self.early_stopping)
        LOGGER.info("%s", msg)

    @staticmethod
    def __mkdir(directory):
        """Create directory if not exist.

        :param Path directory: directory name.
        """
        # TODO: in Python 3, replace this workaround with Path.mkdir(exist_ok=True)
        if not directory.exists():
            directory.mkdir()

    def __create_directories(self):
        """Create the different directories to store results, post-processings, ..."""
        self.__mkdir(self.directory)
        post = self.directory / POST_DIRECTORY
        self.__mkdir(post)
        postoptim = self.directory / POSTOPTIM_DIRECTORY
        self.__mkdir(postoptim)
        poststudy = self.directory / POSTSTUDY_DIRECTORY
        self.__mkdir(poststudy)
        postscal = self.directory / POSTSCAL_DIRECTORY
        self.__mkdir(postscal)
        results = self.directory / RESULTS_DIRECTORY
        self.__mkdir(results)
        msg = MultiLineString()
        msg.indent()
        msg.add("Create directories")
        msg.indent()
        msg.add("Working directory: {}", self.directory)
        msg.add("Post-processing: {}", post)
        msg.add("Optimization history view: {}", postoptim)
        msg.add("Scalability views: {}", poststudy)
        msg.add("Dependency matrices: {}", postscal)
        msg.add("Results: {}", results)
        LOGGER.info("%s", msg)

    def add_discipline(self, data):
        """This method adds a disciplinary dataset from a dataset.

        :param Dataset data: dataset provided as a dataset.
        """
        self._group_dep[data.name] = {}
        self._all_data = data.get_all_data()
        self.datasets.append(data)
        for output_name in data.get_names(data.OUTPUT_GROUP):
            self.set_fill_factor(data.name, output_name, self._default_fill_factor)
        inputs = ", ".join(
            [
                "{}({})".format(name, data.sizes[name])
                for name in data.get_names(data.INPUT_GROUP)
            ]
        )
        outputs = ", ".join(
            [
                "{}({})".format(name, data.sizes[name])
                for name in data.get_names(data.OUTPUT_GROUP)
            ]
        )
        msg = MultiLineString()
        msg.add("Add scalable discipline # {}", len(self.datasets))
        msg.indent()
        msg.add("Name: {}", data.name)
        msg.add("Inputs: {}", inputs)
        msg.add("Outputs: {}", outputs)
        msg.add("Built from {}", len(data))
        LOGGER.info("%s", msg)

    @property
    def disciplines_names(self):
        """Get discipline names.

        :return: list of discipline names
        :rtype: list(str)
        """
        disc_names = [discipline.name for discipline in self.datasets]
        return disc_names

    def set_input_output_dependency(self, discipline, output, inputs):
        """Set the dependency between an output and a set of inputs for a given
        discipline.

        :param str discipline: name of the discipline
        :param str output: name of the output
        :param list(str) inputs: list of inputs names
        """
        self.__check_discipline(discipline)
        self.__check_output(discipline, output)
        self.__check_inputs(discipline, inputs)
        self._group_dep[discipline][output] = inputs

    def set_fill_factor(self, discipline, output, fill_factor):
        """
        :param str discipline: name of the discipline
        :param str output: name of the output function
        :param float fill_factor: fill factor
        """
        self.__check_discipline(discipline)
        self.__check_output(discipline, output)
        self.__check_fill_factor(fill_factor)
        if discipline not in self._fill_factor:
            self._fill_factor[discipline] = {}
        self._fill_factor[discipline][output] = fill_factor

    def __check_discipline(self, discipline):
        """Check if discipline is a string comprised in the list of disciplines
        names."""
        if not isinstance(discipline, string_types):
            raise TypeError("The argument discipline should be a string")
        disciplines_names = self.disciplines_names
        if discipline not in disciplines_names:
            raise ValueError(
                "The argument discipline should be a string comprised in the list %s",
                disciplines_names,
            )

    def __check_output(self, discipline, varname):
        """Check if a variable is an output of a given discipline."""
        self.__check_discipline(discipline)
        if not isinstance(varname, string_types):
            raise TypeError("{} is not a string.".format(varname))
        outputs_names = next(
            dataset.get_names(dataset.OUTPUT_GROUP)
            for dataset in self.datasets
            if dataset.name == discipline
        )
        if varname not in outputs_names:
            raise ValueError(
                "'{}' is not an output of {}; available outputs are: {}".format(
                    varname, discipline, outputs_names
                )
            )

    def __check_inputs(self, discipline, inputs):
        """Check if inputs is a list of inputs of discipline."""
        self.__check_discipline(discipline)
        if not isinstance(inputs, list):
            raise TypeError("The argument 'inputs' must be a list of string.")
        inputs_names = next(
            dataset.get_names(dataset.INPUT_GROUP)
            for dataset in self.datasets
            if dataset.name == discipline
        )
        for inpt in inputs:
            if not isinstance(inpt, string_types):
                raise TypeError("{} is not a string.".format(inpt))
            if inpt not in inputs_names:
                raise ValueError(
                    "'{}' is not a discipline input; available inputs are: {}".format(
                        inpt, inputs_names
                    )
                )

    def __check_fill_factor(self, fill_factor):
        """Check if fill factor is a proportion or a number equal to -1.

        :param float fill_factor: a proportion or -1
        """
        try:
            self.__check_proportion(fill_factor)
        except ValueError:
            if fill_factor != -1:
                raise TypeError(
                    "Fill factor should be a float number comprised in 0 and 1 "
                    "or a number equal to -1."
                )

    @staticmethod
    def __check_proportion(proportion):
        """Check if a proportion is a float number comprised in 0 and 1.

        :param float proportion: proportion comprised in 0 and 1.
        """
        if not isinstance(proportion, numbers.Number):
            raise TypeError(
                "A proportion should be a float number comprised in 0 and 1."
            )
        if not 0 <= proportion <= 1:
            raise ValueError(
                "A proportion should be a float number comprised in 0 and 1."
            )

    def add_optimization_strategy(
        self,
        algo,
        max_iter,
        formulation="DisciplinaryOpt",
        algo_options=None,
        formulation_options=None,
        top_level_diff="auto",
    ):
        """Add both optimization algorithm and MDO formulation, as well as their
        options.

        :param str algo: name of the optimization algorithm.
        :param int max_iter: maximum number of iterations
            for the optimization algorithm.
        :param str formulation: name of the MDO formulation.
            Default: 'DisciplinaryOpt'.
        :param dict algo_options: options of the optimization algorithm.
        :param dict formulation_options: options of the MDO formulation.
        :param str top_level_diff: differentiation method
            for the top level disciplines. Default: 'auto'.
        """
        self.algorithms.append(algo)
        if algo_options is None:
            algo_options = {}
        else:
            if not isinstance(algo_options, dict):
                raise TypeError("algo_options must be a dictionary.")
        algo_options.update({"max_iter": max_iter})
        self.algorithms_options.append(algo_options)
        self.formulations.append(formulation)
        self.formulations_options.append(formulation_options)
        self.top_level_diff.append(top_level_diff)
        if algo_options is not None:
            algo_options = ", ".join(
                ["{}({})".format(name, value) for name, value in algo_options.items()]
            )
        if formulation_options is not None:
            formulation_options = ", ".join(
                [
                    "{}({})".format(name, value)
                    for name, value in formulation_options.items()
                ]
            )
        msg = MultiLineString()
        msg.add("Add optimization strategy # {}", len(self.formulations))
        msg.indent()
        msg.add("Algorithm: {}", algo)
        msg.add("Algorithm options: {}", algo_options)
        msg.add("Formulation: {}", formulation)
        msg.add("Formulation options: {}", formulation_options)
        LOGGER.info("%s", msg)

    def add_scaling_strategies(
        self,
        design_size=None,
        coupling_size=None,
        eq_cstr_size=None,
        ineq_cstr_size=None,
        variables=None,
    ):
        """Add different scaling strategies.

        :param design_size: size of the design variables. Default: None.
        :type design_size: int or list(int)
        :param coupling_size: size of the coupling variables. Default: None.
        :type coupling_size: int or list(int)
        :param eq_cstr_size: size of the equality constraints. Default: None.
        :type eq_cstr_size: int or list(int)
        :param ineq_cstr_size: size of the inequality constraints.
            Default: None.
        :type ineq_cstr_size: int or list(int)
        """
        n_design = self.__check_varsizes_type(design_size)
        n_coupling = self.__check_varsizes_type(coupling_size)
        n_eq = self.__check_varsizes_type(eq_cstr_size)
        n_ineq = self.__check_varsizes_type(ineq_cstr_size)
        n_var = self.__check_varsizes_type(variables)
        n_scaling = max(n_design, n_coupling, n_eq, n_ineq, n_var)
        self.__check_scaling_consistency(n_design, n_scaling)
        self.__check_scaling_consistency(n_coupling, n_scaling)
        self.__check_scaling_consistency(n_eq, n_scaling)
        self.__check_scaling_consistency(n_ineq, n_scaling)
        design_size = self.__format_scaling(design_size, n_scaling)
        coupling_size = self.__format_scaling(coupling_size, n_scaling)
        eq_cstr_size = self.__format_scaling(eq_cstr_size, n_scaling)
        ineq_cstr_size = self.__format_scaling(ineq_cstr_size, n_scaling)
        for idx in range(n_scaling):
            var_scaling = {}
            self.__update_var_scaling(
                var_scaling, design_size[idx], self.design_variables
            )
            self.__update_var_scaling(
                var_scaling, coupling_size[idx], self.coupling_variables
            )
            self.__update_var_scaling(
                var_scaling, eq_cstr_size[idx], self.eq_constraints
            )
            self.__update_var_scaling(
                var_scaling, ineq_cstr_size[idx], self.ineq_constraints
            )
            scaling = {
                "design_variables": design_size[idx],
                "coupling_variables": coupling_size[idx],
                "eq_constraint_size": eq_cstr_size[idx],
                "ineq_constraint_size": ineq_cstr_size[idx],
            }
            if variables is not None and variables[0] is not None:
                for varname, value in variables[idx].items():
                    self.__update_var_scaling(var_scaling, value, [varname])
                    scaling[varname] = value
            else:
                variables = [None] * n_scaling
            self.var_scalings.append(var_scaling)
            self.scalings.append(scaling)
        msg = MultiLineString()
        msg.add("Add scaling strategies")
        msg.indent()
        msg.add("Number of strategies: {}", n_scaling)
        for idx in range(n_scaling):
            if variables[idx] is not None:
                var_str = ", ".join(
                    [
                        "{}({})".format(name, size)
                        for name, size in variables[idx].items()
                    ]
                )
            else:
                var_str = None
            msg.add("Strategy # {}", idx + 1)
            msg.indent()
            msg.add("Design variables: {}", design_size[idx])
            msg.add("Coupling variables: {}", coupling_size[idx])
            msg.add("Equality constraints: {}", eq_cstr_size[idx])
            msg.add("Inequality constraints: {}", ineq_cstr_size[idx])
            msg.add("Variables: {}", var_str)
            msg.dedent()
        LOGGER.info("%s", msg)

    @staticmethod
    def __format_scaling(size, n_scaling):
        """Convert a scaling size in a list of integers whose length is equal to the
        number of scalings.

        :param size: size(s) of a given variable
        :type size: int or list(int)
        :param int n_scaling: number of scalings
        :return: formatted sizes
        """
        formatted_sizes = size
        if isinstance(formatted_sizes, int) or formatted_sizes is None:
            formatted_sizes = [formatted_sizes]
        if len(formatted_sizes) == 1:
            formatted_sizes = formatted_sizes * n_scaling
        return formatted_sizes

    @staticmethod
    def __update_var_scaling(scaling, size, varnames):
        """Update a scaling dictionary for a given list of variables and a given size.

        :param dict scaling: scaling dictionary whose keys are variable names
            and values are dictionary with scaling properties,
            e.g. {'size': val}
        :param int size: size of the variable
        :param list(str) varnames: list of variable names
        """
        if size is not None:
            scaling.update({varname: size for varname in varnames})

    @staticmethod
    def __check_scaling_consistency(n_var_scaling, n_scaling):
        """Check that for the different types of variables, the number of scalings is
        the same or equal to 1.

        :param int n_var_scaling: number of scalings
        :param int n_scaling: expected number of scalings
        """
        assert n_var_scaling in (n_scaling, 1)

    @staticmethod
    def __check_varsizes_type(varsizes):
        """Check the type of scaling sizes. Integer, list of integers or None is
        expected. Return the number of scalings.

        :return: length of scalings
        """
        length = 1
        if varsizes is not None:
            if isinstance(varsizes, list):
                for size in varsizes:
                    if isinstance(size, dict):
                        for _, value in size.items():
                            assert isinstance(value, int)
                    else:
                        assert isinstance(size, int)
                length = len(varsizes)
            else:
                assert isinstance(varsizes, int)
                length = 1
        return length

    def execute(self, n_replicates=1):
        """Execute the scalability study, one or several times to take into account the
        random features of the scalable problems.

        :param int n_replicates: number of times the scalability study
            is repeated. Default: 1.
        """
        plural = "s" if n_replicates > 1 else ""
        LOGGER.info("Execute scalability study %s time%s", n_replicates, plural)
        if not self.formulations and not self.algorithms:
            raise ValueError(
                "A scalable study needs at least 1 optimization strategy, "
                "defined by a mandatory optimization algorithm "
                "and optional optimization algorithm and options"
            )
        counter = "Formulation: {} - Algo: {} - Scaling: {}/{} - Replicate: {}/{}"
        n_scal_strategies = len(self.var_scalings)
        n_opt_strategies = len(self.algorithms)
        for opt_index in range(n_opt_strategies):
            algo = self.algorithms[opt_index]
            formulation = self.formulations[opt_index]
            for scal_index in range(n_scal_strategies):
                scaling = self.var_scalings[scal_index]
                for replicate in range(1, n_replicates + 1):
                    msg = MultiLineString()
                    msg.indent()
                    data = (
                        formulation,
                        algo,
                        scal_index + 1,
                        n_scal_strategies,
                        replicate,
                        n_replicates,
                    )
                    msg.add(counter, *data)
                    LOGGER.info("%s", msg)
                    msg = MultiLineString()
                    msg.indent()
                    msg.indent()
                    msg.add("Create scalable problem")
                    problem = self.__create_scalable_problem(scaling, replicate)
                    path = self.__dep_mat_path(algo, formulation, scal_index, replicate)
                    directory = path.stem
                    self.__mkdir(path)
                    msg.add("Save dependency matrices in {}", path)
                    problem.plot_dependencies(True, False, str(path))
                    msg.add("Create MDO Scenario")
                    root_logger = logging.getLogger()
                    saved_level = root_logger.level
                    root_logger.setLevel(logging.WARNING)
                    self.__create_scenario(problem, formulation, opt_index)
                    msg.add("Execute MDO Scenario")
                    formulation_options = self.formulations_options[opt_index]
                    algo_options = self.__execute_scenario(problem, algo, opt_index)
                    root_logger.setLevel(saved_level)
                    path = self.__optview_path(algo, formulation, scal_index, replicate)
                    msg.add("Save optim history view in {}", path)
                    fpath = str(path) + "/"
                    problem.scenario.post_process(
                        "OptHistoryView", save=True, show=False, file_path=fpath
                    )
                    result = ScalabilityResult(directory, scal_index + 1, replicate)
                    self.results.append(result)
                    statistics = self.__get_statistics(problem, scaling)
                    result.get(
                        algo=algo,
                        algo_options=algo_options,
                        formulation=formulation,
                        formulation_options=formulation_options,
                        scaling=scaling,
                        disc_names=problem.disciplines,
                        output_names=problem.outputs,
                        **statistics
                    )
                    fpath = result.get_file_path(self.directory)
                    msg.add("Save statistics in {}", fpath)
                    result.save(str(self.directory))
                    LOGGER.debug("%s", msg)
        return self.results

    def __get_statistics(self, problem, scaling):
        """Get statistics from an executed scalable problem.

        :param ScalableProblem problem: scalable problem.
        :param dict scaling: variables scaling.
        """
        statistics = {}
        stopidx, n_iter = self.__get_stop_index(problem)
        ratio = float(stopidx) / n_iter
        n_calls = {disc: n_calls * ratio for disc, n_calls in problem.n_calls.items()}
        statistics["n_calls"] = n_calls
        tmp = problem.n_calls_linearize
        n_calls_linearize = {disc: ncl * ratio for disc, ncl in tmp.items()}
        statistics["n_calls_linearize"] = n_calls_linearize
        tmp = problem.n_calls_top_level
        n_calls_tl = {disc: n_calls * ratio for disc, n_calls in tmp.items()}
        statistics["n_calls_top_level"] = n_calls_tl
        tmp = problem.n_calls_linearize_top_level
        n_calls_linearize_tl = {disc: ncltl * ratio for disc, ncltl in tmp.items()}
        statistics["n_calls_linearize_top_level"] = n_calls_linearize_tl
        statistics["exec_time"] = problem.exec_time() * ratio
        statistics["status"] = problem.status
        statistics["is_feasible"] = problem.is_feasible

        inputs = problem.inputs
        outputs = problem.outputs
        disc_varnames = {
            disc: inputs[disc] + outputs[disc] for disc in problem.disciplines
        }
        sizes = problem.varsizes
        statistics["new_varsizes"] = {
            disc: {name: scaling.get(name, sizes[name]) for name in disc_varnames[disc]}
            for disc in problem.disciplines
        }
        statistics["old_varsizes"] = problem.varsizes
        return statistics

    def __dep_mat_path(self, algo, formulation, id_scaling, replicate):
        """Path to the directory containing the dependency matrices files.

        :param str algo: algo name.
        :param str formulation: formulation name.
        :param int id_scaling: scaling index.
        :param int replicate: replicate number.
        """
        varnames = [algo, formulation, id_scaling + 1, replicate]
        name = "_".join([self.prefix] + [str(var) for var in varnames])
        if name[0] == "_":
            name = name[1:]
        path = self.directory / POSTSCAL_DIRECTORY / name
        return path

    def __optview_path(self, algo, formulation, id_scaling, replicate):
        """Path to the directory containing the dependency matrices files.

        :param str algo: algo name.
        :param str formulation: formulation name.
        :param int id_scaling: scaling index.
        :param int replicate: replicate number.
        """
        directory = Path(self.prefix + "_" + algo + "_" + formulation)
        path = self.directory / POSTOPTIM_DIRECTORY / directory
        self.__mkdir(path)
        path = path / Path("scaling_{}".format(id_scaling + 1))
        self.__mkdir(path)
        path = path / Path("replicate_{}".format(replicate))
        self.__mkdir(path)
        return path

    def __create_scalable_problem(self, scaling, seed):
        """Create a scalable problem.

        :param dict scaling: scaling.
        :param int seed: seed for random features.
        """
        problem = ScalableProblem(
            self.datasets,
            self.design_variables,
            self.objective,
            self.eq_constraints,
            self.ineq_constraints,
            self.maximize_objective,
            scaling,
            fill_factor=self._fill_factor,
            seed=seed,
            group_dep=self._group_dep,
            force_input_dependency=True,
            allow_unused_inputs=False,
        )
        return problem

    def __create_scenario(self, problem, formulation, opt_index):
        """Create scenario for a given formulation.

        :param ScalableProblem problem: scalable problem.
        :param str formulation: MDO formulation name.
        :param int opt_index: optimization strategy index.
        """
        form_opt = self.formulations_options[opt_index]
        if not isinstance(form_opt, dict):
            formulation_options = {}
        else:
            formulation_options = form_opt
        problem.create_scenario(
            formulation,
            "MDO",
            self.start_at_equilibrium,
            self.active_probability,
            self.feasibility_level,
            **formulation_options
        )

    def __execute_scenario(self, problem, algo, opt_index):
        """Execute scenario.

        :param ScalableProblem problem: scalable problem.
        :param str algo: optimization algorithm name.
        :param int opt_index: optimization strategy index.
        """
        top_level_disciplines = problem.scenario.formulation.get_top_level_disc()
        for disc in top_level_disciplines:
            disc.linearization_mode = self.top_level_diff[opt_index]
        algo_options = deepcopy(self.algorithms_options[opt_index])
        max_iter = algo_options["max_iter"]
        del algo_options["max_iter"]
        problem.scenario.execute(
            {"algo": algo, "max_iter": max_iter, "algo_options": algo_options}
        )
        return algo_options

    def __get_stop_index(self, problem):
        """Get stop index from a database.

        :param ScalableProblem problem: scalable problem
        :return: stop index, database length
        :rtype: int, int
        """
        database = problem.scenario.formulation.opt_problem.database
        n_iter = len(database)
        if self.early_stopping:
            y_prev = inf
            stopidx = 0
            for _, value in database.items():
                pbm = problem.scenario.formulation.opt_problem
                if y_prev == inf:
                    diff = inf
                else:
                    diff = abs(y_prev - value[pbm.get_objective_name()])
                    diff /= abs(y_prev)
                if diff < 1e-6:
                    break
                y_prev = value[pbm.get_objective_name()]
                stopidx += 1
        else:
            stopidx = n_iter
        return stopidx, n_iter



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Base class for optimization history post-processing."""
from __future__ import division, unicode_literals

import inspect
from os.path import abspath, dirname, exists, join
from typing import Dict, List, Optional, Sequence, Tuple, Union

import six
from custom_inherit import DocInheritMeta
from matplotlib.figure import Figure

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.grammar import InvalidDataException
from gemseo.core.json_grammar import JSONGrammar
from gemseo.post.dataset.dataset_plot import DatasetPlot
from gemseo.utils.file_path_manager import FilePathManager, FileType
from gemseo.utils.matplotlib_figure import save_show_figure
from gemseo.utils.py23_compat import OrderedDict, Path
from gemseo.utils.source_parsing import get_options_doc

OptPostProcessorOptionType = Union[int, float, str, bool, Sequence[str]]
PlotOutputType = List[
    Tuple[Optional[str], Union[Figure, DatasetPlot], Optional[Dict[str, Sequence[str]]]]
]


@six.add_metaclass(DocInheritMeta(abstract_base_class=True))
class OptPostProcessor(object):
    """Abstract class for optimization post-processing methods.

    Attributes:
        opt_problem (OptimizationProblem): The optimization problem.
        database (Database): The database generated by the optimization problem.
        out_data_dict (Dict[Any,Any]): The data dict
            to eventually rebuild the plot in another framework.
    """

    def __init__(
        self,
        opt_problem,  # type: OptimizationProblem
    ):  # type: (...) -> None
        """
        Args:
            opt_problem: The optimization problem to be post-processed.

        Raises:
            ValueError: If the JSON grammar file
                for the options of the post-processor does not exist.
        """
        self.opt_problem = opt_problem
        self.database = opt_problem.database
        comp_dir = abspath(dirname(inspect.getfile(OptPostProcessor)))
        schema_file = join(comp_dir, "OptPostProcessor.json")
        self.opt_grammar = JSONGrammar(
            "OptPostProcessor",
            schema_file=schema_file,
            descriptions=get_options_doc(self.execute),
        )
        cls_name = self.__class__.__name__
        name = cls_name + "_options"
        f_class = inspect.getfile(self.__class__)
        comp_dir = abspath(dirname(f_class))
        schema_file = join(comp_dir, "{}.json".format(name))
        if not exists(schema_file):
            schema_file = join(comp_dir, "options", "{}.json".format(name))
        if not exists(schema_file):
            raise ValueError(
                "Options grammar for optimization post-processor does not exist, "
                "expected: {} or {}".format(schema_file, join(comp_dir, name + ".json"))
            )

        descriptions = {}

        if hasattr(self.__class__, "_run"):
            descriptions.update(get_options_doc(self.__class__._run))

        if hasattr(self.__class__, "_plot"):
            descriptions.update(get_options_doc(self.__class__._plot))

        self.opt_grammar.update_from(
            JSONGrammar(name, schema_file=schema_file, descriptions=descriptions)
        )

        # the data dict to eventually rebuild the plot in another framework
        self.out_data_dict = {}
        default_file_name = FilePathManager.to_snake_case(self.__class__.__name__)
        self.__file_path_manager = FilePathManager(FileType.FIGURE, default_file_name)
        self.__output_files = []
        self.__figures = OrderedDict()
        self.__nameless_figure_counter = 0

    @property
    def figures(self):  # type: (...) -> Dict[str,Figure]
        """The Matplotlib figures indexed by a name, or the nameless figure counter."""
        return self.__figures

    @property
    def output_files(self):  # type: (...) -> List[str]
        """The paths to the output files."""
        return self.__output_files

    def _add_figure(
        self,
        figure,  # type: Figure
        file_name=None,  # type: Optional[str]
    ):  # type: (...) -> None
        """Add a figure.

        Args:
            figure: The figure to be added.
            file_name: The default name of the file to save the figure.
                If None, use the nameless figure counter.
        """
        if file_name is None:
            self.__nameless_figure_counter += 1
            file_name = str(self.__nameless_figure_counter)

        self.__figures[file_name] = figure

    def execute(
        self,
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_extension=None,  # type: Optional[str]
        **options  # type: OptPostProcessorOptionType
    ):  # type: (...) -> Dict[str,Figure]
        """Post-process the optimization problem.

        Args:
            save: If True, save the figure.
            show: If True, display the figure.
            file_path: The path of the file to save the figures.
                If the extension is missing, use ``file_extension``.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_extension: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.
            **options: The options of the post-processor.

        Returns:
            The figure, to be customized if not closed.

        Raises:
            ValueError: If the `opt_problem.database` is empty.
        """
        # convert file_path to string before grammar-based options checking
        if isinstance(file_path, Path):
            file_path_to_be_checked = str(file_path)
        else:
            file_path_to_be_checked = file_path

        if isinstance(directory_path, Path):
            directory_path_to_be_checked = str(directory_path)
        else:
            directory_path_to_be_checked = directory_path

        if file_path is not None:
            file_path = Path(file_path)

        if directory_path is not None:
            directory_path = Path(directory_path)

        self.check_options(
            save=save,
            show=show,
            file_path=file_path_to_be_checked,
            file_name=file_name,
            directory_path=directory_path_to_be_checked,
            file_extension=file_extension,
            **options
        )
        if not self.opt_problem.database:
            raise ValueError(
                "Optimization problem was not solved, "
                "cannot run post processing {}".format(self.__class__.__name__)
            )

        self.__figures = self._run(
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_extension=file_extension,
            directory_path=directory_path,
            **options
        )
        return self.__figures

    def check_options(
        self, **options  # type: OptPostProcessorOptionType
    ):  # type: (...) -> None
        """Check the options of the post-processor.

        Args:
            **options: The options of the post-processor.

        Raises:
            InvalidDataException: If an option is invalid according to the grammar.
        """
        try:
            self.opt_grammar.load_data(options)
        except InvalidDataException:
            raise InvalidDataException(
                "Invalid options for post-processor {}; "
                "got: {}".format(self.__class__.__name__, options)
            )

    def _run(
        self,
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Path]
        directory_path=None,  # type: Optional[Path]
        file_name=None,  # type: Optional[str]
        file_extension=None,  # type: Optional[str]
        **options  # type: OptPostProcessorOptionType
    ):  # type: (...) -> Dict[str,Figure]
        """Run the post-processor.

        Args:
            save: If True, save the figure.
            show: If True, display the figure.
            file_path: The path of the file to save the figures.
                If the extension is missing, use ``file_extension``.
                If None,
                create a file path
                from ``directory_path``, ``file_name`` and ``file_extension``.
            directory_path: The path of the directory to save the figures.
                If None, use the current working directory.
            file_name: The name of the file to save the figures.
                If None, use a default one generated by the post-processing.
            file_extension: A file extension, e.g. 'png', 'pdf', 'svg', ...
                If None, use a default file extension.
            **options: The options of the post-processor.

        Returns:
            The figures resulting from the post-processing of the optimization problem.
        """
        self._plot(**options)

        file_path = self.__file_path_manager.create_file_path(
            file_path=file_path,
            directory_path=directory_path,
            file_name=file_name,
            file_extension=file_extension,
        )

        for figure_name, figure in self.__figures.items():
            if save:
                if len(self.__figures) > 1:
                    fig_file_path = self.__file_path_manager.add_suffix(
                        file_path, figure_name
                    )
                else:
                    fig_file_path = file_path

                self.__output_files.append(str(fig_file_path))

            else:
                fig_file_path = None

            save_show_figure(figure, show, fig_file_path)

        return self.__figures

    def _plot(
        self, **options  # type: OptPostProcessorOptionType
    ):  # type: (...) -> None
        """Create the figures.

        Args:
            **options: The post-processor options.
        """
        raise NotImplementedError()

    def _generate_x_names(self):  # type: (...)-> List[str]
        """Create the design variables names for the plot.

        Returns:
            The design variables names.
        """

        design_variables = self.opt_problem.get_design_variable_names()
        x_names = []
        for d_v in design_variables:
            dv_size = self.opt_problem.design_space.variables_sizes[d_v]
            if dv_size == 1:
                x_names.append(d_v)
            else:
                for k in range(dv_size):
                    x_names.append("{}_{}".format(d_v, k))
        return x_names



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Remi Lafage
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Abstraction for workflow
************************
"""
from __future__ import division, unicode_literals

import logging
from uuid import uuid4

from gemseo.core.discipline import MDODiscipline
from gemseo.utils.py23_compat import OrderedDict  # automatically dict from py36

LOGGER = logging.getLogger(__name__)

STATUS_FAILED = MDODiscipline.STATUS_FAILED
STATUS_DONE = MDODiscipline.STATUS_DONE
STATUS_PENDING = MDODiscipline.STATUS_PENDING
STATUS_RUNNING = MDODiscipline.STATUS_RUNNING


class ExecutionSequence(object):
    """A base class for execution sequences.

    The execution sequence structure is introduced to reflect the main workflow
    implicitly executed by |g| regarding the given scenario/formulation executed. That
    structure allows to identify single executions of a same discipline that may be run
    several times at various stages in the given scenario/formulation.
    """

    START_STR = "["
    END_STR = "]"

    def __init__(self, sequence=None):  # pylint: disable=unused-argument
        # use an OrderedDict to get disc_to_uuids lists ordered regarding
        # a discipline repetitive appearance: useful for testing and debug
        self.uuid = str(uuid4())
        self.uuid_to_disc = OrderedDict()
        self.disc_to_uuids = {}
        self._status = None
        self._enabled = False
        self._parent = None

    def accept(self, visitor):
        """Accept a visitor object (see Visitor pattern). Have to be implemented by
        subclasses.

        :param visitor: a visitor object
        """
        raise NotImplementedError()

    def set_observer(self, obs):
        """Register the given observer object which is intended to be notified via its
        update() method each time an underlying discipline changes its status. To be
        implemented in subclasses.

        :returns: the disciplines list.
        """
        raise NotImplementedError()

    @property
    def status(self):
        """Get status value.

        :returns: the status value (MDODiscipline.STATUS_XXX values).
        """
        return self._status

    @status.setter
    def status(self, status):
        """Set status value.

        :param status: (MDODiscipline.STATUS_XXX values).
        """
        self._status = status

    @property
    def parent(self):
        """Get the containing execution sequence.

        :returns: a composite execution sequence.
        """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """Set the containing execution sequence as parent. self should be included in
        parent.sequence_list.

        :returns: the status value.
        """
        if self not in parent.sequence_list:
            raise RuntimeError(
                "parent " + str(parent) + " do not include child " + str(self)
            )
        self._parent = parent

    def enabled(self):
        """Get activation state.

        :returns: boolean True if enabled.
        """
        return self._enabled

    def enable(self):
        """Set the execution sequence as activated (enabled)."""
        self.status = STATUS_PENDING
        self._enabled = True

    def disable(self):
        """Set the execution sequence as deactivated (disabled)."""
        self._enabled = False

    def _compute_disc_to_uuids(self):
        """
        Update discipline to uuids mapping from uuids to discipline mapping
        Note: a discipline might correspond to several
        AtomicExecutionSeuqence hence might correspond to several uuids.

        """
        self.disc_to_uuids = {}
        for key, value in self.uuid_to_disc.items():
            self.disc_to_uuids.setdefault(value, []).append(key)


class AtomicExecSequence(ExecutionSequence):
    """An execution sequence to represent the single execution of a given discipline."""

    def __init__(self, discipline=None):
        super(AtomicExecSequence, self).__init__(discipline)
        if not isinstance(discipline, MDODiscipline):
            raise Exception(
                "Atomic sequence shall be a discipline"
                + ", got "
                + str(type(discipline))
                + " instead !"
            )
        self.discipline = discipline
        self.uuid_to_disc = {self.uuid: discipline}
        self.disc_to_uuids = {discipline: [self.uuid]}
        self._observer = None

    def __str__(self):
        return self.discipline.name + "(" + str(self.status) + ")"

    def __repr__(self):
        return (
            self.discipline.name + "(" + str(self.status) + ", " + str(self.uuid) + ")"
        )

    def accept(self, visitor):
        """Accept a visitor object (see Visitor pattern)

        :param visitor: a visitor object implementing visit_atomic() method
        """
        visitor.visit_atomic(self)

    def set_observer(self, obs):
        """Register given observer obs to be notified (obs.update()) when discipline
        status changes.

        :param obs: the observe object implementing update() method
        """
        self._observer = obs

    def enable(self):
        """Subscribe to status changes of the discipline (notified via
        update_status())"""
        super(AtomicExecSequence, self).enable()
        self.discipline.add_status_observer(self)

    def disable(self):
        """Unsubscribe from receiving status changes of the discipline."""
        super(AtomicExecSequence, self).disable()
        self.discipline.remove_status_observer(self)

    def get_state_dict(self):
        """Get the dictionary of statuses mapping atom uuid to status.

        :returns: the status
        """

        return {self.uuid: self.status}

    def update_status(self, discipline):
        """
        Update status from given discipline.
        Reflect the status then notifies the parent and the observer if any.
        Note: update_status if discipline status change actually
        compared to current, otherwise do nothing.

        :param discipline: the discipline whose status changed

        """
        if self._enabled and self.status != discipline.status:
            self.status = discipline.status or STATUS_PENDING
            if self.status == STATUS_DONE or self.status == STATUS_FAILED:
                self.disable()
            if self._parent:
                self._parent.update_child_status(self)
            if self._observer:
                self._observer.update(self)

    def force_statuses(self, status):
        """Force the self status and the status of subsequences without notifying the
        parent (as the force_status is called by a parent), but notify the observer is
        status changed.

        :param: status value (see MDODiscipline.STATUS_XXX values)
        """
        old_status = self._status
        self._status = status
        if old_status != status and self._observer:
            self._observer.update(self)


class CompositeExecSequence(ExecutionSequence):
    """A base class for execution sequence made of other execution sequences.

    Intented to be subclassed.
    """

    START_STR = "'"
    END_STR = "'"

    def __init__(self, sequence=None):
        super(CompositeExecSequence, self).__init__(sequence)
        self.sequence_list = []
        self.disciplines = []

    def __str__(self):
        str_out = self.START_STR
        for seq in self.sequence_list:
            str_out += str(seq) + ", "
        str_out += self.END_STR
        return str_out

    def accept(self, visitor):
        """Accept a visitor object (see Visitor pattern) and then make its children
        accept it too.

        :param visitor: a visitor object implementing visit_serial() method
        """
        self._accept(visitor)
        for seq in self.sequence_list:
            seq.accept(visitor)

    def _accept(self, visitor):
        """Accept a visitor object (see Visitor pattern). To be specifically implemented
        by subclasses to call relevant visitor method depending the subclass type.

        :param visitor: a visitor object implementing visit_serial() method
        """
        raise NotImplementedError()

    def set_observer(self, obs):
        """Set observer obs to subsequences. Override super.set_observer()

        :param obs: observer object implementing update() method
        """
        for seq in self.sequence_list:
            seq.set_observer(obs)

    def disable(self):
        """Unsubscribe subsequences from receiving status changes of disciplines."""
        super(CompositeExecSequence, self).disable()
        for seq in self.sequence_list:
            seq.disable()

    def force_statuses(self, status):
        """Force the self status and the status of subsequences.

        params: status value (see MDODiscipline.STATUS_XXX values)
        """
        self.status = status
        for seq in self.sequence_list:
            seq.force_statuses(status)

    def get_state_dict(self):
        """Get the dictionary of statuses mapping atom uuid to status.

        :returns: the status
        """
        state_dict = {}
        for seq in self.sequence_list:
            state_dict.update(seq.get_state_dict())
        return state_dict

    def update_child_status(self, child):
        """Manage status change of child execution sequences. Propagates status change
        to the parent (containing execution sequence)

        :param child: the child execution sequence (contained in sequence_list)
            whose status has changed
        """
        old_status = self.status
        self._update_child_status(child)
        if self._parent and self.status != old_status:
            self._parent.update_child_status(self)

    def _update_child_status(self, child):
        """Handle child execution change. To be implemented in subclasses.

        :param child: the child execution sequence (contained in sequence_list)
            whose status has changed
        """
        raise NotImplementedError()


class ExtendableExecSequence(CompositeExecSequence):
    """A base class for composite execution sequence that are extendable.

    Intented to be subclassed.
    """

    def __init__(self, sequence=None):
        super(ExtendableExecSequence, self).__init__(sequence)
        if sequence is not None:
            self.extend(sequence)

    def extend(self, sequence):
        """Extend the execution sequence with another ExecutionSequence or a discipline.

        :param sequence: another execution sequence or
        """
        seq_class = sequence.__class__
        self_class = self.__class__
        if isinstance(sequence, list):
            # In this case we are initializing the sequence
            # or extending by a list of disciplines
            self._extend_with_disc_list(sequence)
        elif isinstance(sequence, MDODiscipline):
            # Sequence is extended by a single discipline: generate a new
            # uuid
            self._extend_with_disc_list([sequence])
        elif isinstance(sequence, AtomicExecSequence):
            # Sequence is extended by an AtomicSequence:
            # we extend
            self._extend_with_atomic_sequence(sequence)
        elif seq_class != self_class:
            # We extend by a different type of ExecSequence
            # So we append the other sequence as a sub structure
            self._extend_with_diff_sequence_kind(sequence)
        else:
            # We extend by a same type of ExecSequence
            # So we just extend the sequence
            self._extend_with_same_sequence_kind(sequence)
        self._compute_disc_to_uuids()  # refresh disc_to_uuids
        for seq in self.sequence_list:
            seq.parent = self
        return self

    def _extend_with_disc_list(self, sequence):
        """Extend by a list of disciplines.

        :param sequence: a list of MDODiscipline objects
        """
        seq_list = [AtomicExecSequence(disc) for disc in sequence]
        self.sequence_list.extend(seq_list)
        uuids_dict = {atom.uuid: atom.discipline for atom in seq_list}
        self.uuid_to_disc.update(uuids_dict)

    def _extend_with_atomic_sequence(self, sequence):
        """Extend by a list of AtomicExecutionSequence.

        :param sequence: a list of MDODiscipline objects
        """
        self.sequence_list.append(sequence)
        self.uuid_to_disc[sequence.uuid] = sequence

    def _extend_with_same_sequence_kind(self, sequence):
        """Extend by another ExecutionSequence of same type.

        :param sequence: an ExecutionSequence of same type as self
        """
        self.sequence_list.extend(sequence.sequence_list)
        self.uuid_to_disc.update(sequence.uuid_to_disc)

    def _extend_with_diff_sequence_kind(self, sequence):
        """Extend by another ExecutionSequence of different type.

        :param sequence: an ExecutionSequence of type different from self's one
        """
        self.sequence_list.append(sequence)
        self.uuid_to_disc.update(sequence.uuid_to_disc)

    def _update_child_status(self, child):
        """Manage status change of child execution sequences. Done status management is
        handled in subclasses.

        :param child: the child execution sequence (contained in sequence_list)
            whose status has changed
        """
        if child.status == STATUS_FAILED:
            self.status = STATUS_FAILED
        elif child.status == STATUS_DONE:
            self._update_child_done_status(child)
        else:
            self.status = child.status

    def _update_child_done_status(self, child):
        """Handle done status of child execution sequences. To be implemented in
        subclasses.

        :param child: the child execution sequence (contained in sequence_list)
            whose status has changed
        """
        raise NotImplementedError()


class SerialExecSequence(ExtendableExecSequence):
    """A class to describe a serial execution of disciplines."""

    START_STR = "["
    END_STR = "]"

    def __init__(self, sequence=None):
        super(SerialExecSequence, self).__init__(sequence)
        self.exec_index = None

    def _accept(self, visitor):
        """Accept a visitor object (see Visitor pattern)

        :param visitor: a visitor object implementing visit_serial() method
        """
        visitor.visit_serial(self)

    def enable(self):
        """Activate first child execution sequence."""
        super(SerialExecSequence, self).enable()
        self.exec_index = 0
        if self.sequence_list:
            self.sequence_list[self.exec_index].enable()
        else:
            raise Exception("Serial execution is empty")

    def _update_child_done_status(self, child):
        """Activate next child to given child execution sequence. Disable itself when
        all children done.

        :param child: the child execution sequence in done state.
        """
        if child.status == STATUS_DONE:
            child.disable()
            self.exec_index += 1
            if self.exec_index < len(self.sequence_list):
                self.sequence_list[self.exec_index].enable()
            else:  # last seq done
                self.status = STATUS_DONE
                self.disable()


class ParallelExecSequence(ExtendableExecSequence):
    """A class to describe a parallel execution of disciplines."""

    START_STR = "("
    END_STR = ")"

    def _accept(self, visitor):
        """Accept a visitor object (see Visitor pattern)

        :param visitor: a visitor object implementing visit_parallel() method
        """
        visitor.visit_parallel(self)

    def enable(self):
        """Activate all child execution sequences."""
        super(ParallelExecSequence, self).enable()
        for seq in self.sequence_list:
            seq.enable()

    def _update_child_done_status(self, child):  # pylint: disable=unused-argument
        """Disable itself when all children done.

        :param child: the child execution sequence in done state.
        """
        all_done = True
        for seq in self.sequence_list:
            all_done = all_done and (seq.status == STATUS_DONE)
        if all_done:
            self.status = STATUS_DONE
            self.disable()


class LoopExecSequence(CompositeExecSequence):
    """A class to describe a loop with a controller discipline and an execution_sequence
    as iterate."""

    START_STR = "{"
    END_STR = "}"

    def __init__(self, controller, sequence):
        if isinstance(controller, AtomicExecSequence):
            control = controller
        elif not isinstance(controller, MDODiscipline):
            raise Exception(
                "Controller of a loop shall be a discipline"
                + ", got "
                + str(type(controller))
                + " instead !"
            )
        else:
            control = AtomicExecSequence(controller)
        if not isinstance(sequence, CompositeExecSequence):
            raise Exception(
                "Sequence of a loop shall be a composite execution sequence"
                + ", got "
                + str(type(sequence))
                + " instead !"
            )
        super(LoopExecSequence, self).__init__()
        self.sequence_list = [control, sequence]
        self.atom_controller = control
        self.atom_controller.parent = self
        self.iteration_sequence = sequence
        self.iteration_sequence.parent = self
        self.uuid_to_disc.update(sequence.uuid_to_disc)
        self.uuid_to_disc[self.atom_controller.uuid] = controller
        self._compute_disc_to_uuids()
        self.iteration_count = 0

    def _accept(self, visitor):
        """Accept a visitor object (see Visitor pattern)

        :param visitor: a visitor object implementing visit_loop() method
        """
        visitor.visit_loop(self)

    def enable(self):
        """Active controller execution sequence."""
        super(LoopExecSequence, self).enable()
        self.atom_controller.enable()
        self.iteration_count = 0

    def _update_child_status(self, child):
        """Activate iteration successively regarding controller status. Count iterations
        regarding iteration_sequence status.

        :param child: the child execution sequence in done state.
        """
        self.status = self.atom_controller.status
        if child == self.atom_controller:
            if self.status == STATUS_RUNNING:
                if not self.iteration_sequence.enabled():
                    self.iteration_sequence.enable()
            elif self.status == STATUS_DONE:
                self.disable()
                self.force_statuses(STATUS_DONE)
        if child == self.iteration_sequence:
            if child.status == STATUS_DONE:
                self.iteration_count += 1
                self.iteration_sequence.enable()
        if child.status == STATUS_FAILED:
            self.status = STATUS_FAILED


class ExecutionSequenceFactory(object):
    """A factory class for ExecutionSequence objects.

    Allow to create AtomicExecutionSequence, SerialExecutionSequence,
    ParallelExecutionSequence and LoopExecutionSequence. Main |g| workflow is intended
    to be expressed with those four ExecutionSequence types
    """

    @staticmethod
    def atom(discipline):
        """Returns a structure representing the execution of a discipline. This function
        is intended to be called by MDOFormulation.get_expected_workflow methods.

        :param discipline: a discipline
        :returns: the structure used within XDSM workflow representation
        """
        return AtomicExecSequence(discipline)

    @staticmethod
    def serial(sequence=None):
        """Returns a structure representing the serial execution of the given
        disciplines. This function is intended to be called by
        MDOFormulation.get_expected_workflow methods.

        :param sequence: any number of discipline
            or the return value of a serial, parallel or loop call
        :returns: a serial execution sequence
        """
        return SerialExecSequence(sequence)

    @staticmethod
    def parallel(sequence=None):
        """Returns a structure representing the parallel execution of the given
        disciplines. This function is intended to be called by
        MDOFormulation.get_expected_workflow methods.

        :param sequence: any number of discipline or
            the return value of a serial, parallel or loop call
        :returns: a parallel execution sequence
        """
        return ParallelExecSequence(sequence)

    @staticmethod
    def loop(control, composite_sequence):
        """Returns a structure representing a loop execution of a This function is
        intended to be called by MDOFormulation.get_expected_workflow methods.

        :param control: the discipline object, controller of the loop
        :param composite_sequence: any number of discipline
            or the return value of a serial, parallel or loop call
        :returns: a loop execution sequence
        """
        return LoopExecSequence(control, composite_sequence)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Rules and checks for disciplines inputs/outputs validation."""

from __future__ import division, unicode_literals

import logging

from numpy import zeros

LOGGER = logging.getLogger(__name__)


class AbstractGrammar(object):
    """Abstract Grammar : defines the abstraction for a Grammar.

    A grammar subclass instance stores the input or output data
    types and structure an MDODiscipline.
    It is able to check the inputs and outputs against predefined types.
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "grammar name: {}".format(self.name)

    def load_data(self, data, raise_exception=True):
        """Loads the data dictionary in the grammar and checks it against self
        properties.

        :param data: the input data
        :param raise_exception: if False, no exception is raised
             when data is invalid (Default value = True)
        """
        raise NotImplementedError()

    def get_data_names(self):
        """Returns the list of data names.

        :returns: the data names alphabetically sorted
        """
        raise NotImplementedError()

    def update_from(self, input_grammar):
        """Adds properties coming from another grammar.

        :param input_grammar: the grammar to take inputs from
        """
        raise NotImplementedError()

    def update_from_if_not_in(self, input_grammar, exclude_grammar):
        """Adds properties coming from input_grammar if they are not in exclude_grammar.

        :param input_grammar: the grammar to take inputs from
        :param exclude_grammar: exclusion grammar
        """
        raise NotImplementedError()

    def is_data_name_existing(self, data_name):
        """Checks if data_name is present in grammar.

        :param data_name: the data name
        :returns: True if data is in grammar
        """
        raise NotImplementedError()

    def is_all_data_names_existing(self, data_names):
        """Checks if data_names are present in grammar.

        :param data_names: the data names list
        :returns: True if all data are in grammar
        """
        raise NotImplementedError()

    def clear(self):
        """Clears the data to produce an empty grammar."""
        raise NotImplementedError()

    def to_simple_grammar(self):
        """Converts to the base SimpleGrammar type.

        :returns: a SimpleGrammar instance equivalent to self
        """
        raise NotImplementedError()

    def initialize_from_data_names(self, data_names):
        """Initializes a Grammar from a list of data. All data of the grammar will be
        set as arrays.

        :param data_names: a data names list
        """
        data = zeros(1)
        typical_data_dict = {k: data for k in data_names}
        self.initialize_from_base_dict(typical_data_dict)

    def initialize_from_base_dict(self, typical_data_dict):
        """Initialize the grammar with types and names from a typical data entry.

        :param typical_data_dict: a data dictionary
            keys are used as data names
            values are used to detect the data types
        """
        raise NotImplementedError()

    @staticmethod
    def _get_update_error_msg(grammar1, grammar2, grammar3=None):
        """Create a message for grammar update error.

        Args:
            grammar1: A grammar.
            grammar2: A grammar.
            grammar3: A grammar, optional.

        Returns:
            str: The error message.
        """
        msg = "Cannot update grammar {} of type {} with {} of type {}".format(
            grammar1.name,
            grammar1.__class__.__name__,
            grammar2.name,
            grammar2.__class__.__name__,
        )
        if grammar3 is not None:
            msg += " and {} of type {}".format(
                grammar1.name,
                grammar1.__class__.__name__,
            )
        return msg



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains a factory to instantiate a :class:`.AbstractCache` from its
class name.

The class can be internal to |g| or located in an external module whose path is provided
to the constructor. It also provides a list of available cache types and allows you to
test if a cache type is available.
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.cache import AbstractCache
from gemseo.core.factory import Factory

LOGGER = logging.getLogger(__name__)


class CacheFactory(object):
    """This factory instantiates a :class:`.AbstractCache` from its class name.

    The class can be internal to |g| or located in an external module whose path is
    provided to the constructor.
    """

    def __init__(self):
        """Initializes the factory: scans the directories to search for subclasses of
        AbstractCache.

        Searches in "|g|" and gemseo.caches
        """
        self.factory = Factory(AbstractCache, ("gemseo.caches",))

    def create(self, cache_name, **options):
        """Create a cache.

        :param str cache_name: name of the cache (its classname)
        :param options: additional options specific
        :return: cache_name cache

        Examples
        --------
        >>> from gemseo.caches.cache_factory import CacheFactory
        >>> cache = CacheFactory().create('MemoryFullCache', name='my_cache')
         my_cache
        |_ Type: MemoryFullCache
        |_ Input names: None
        |_ Output names: None
        |_ Length: 0
        |_ Tolerance: 0.0
        """
        return self.factory.create(cache_name, **options)

    @property
    def caches(self):
        """Lists the available classes.

        :returns: the list of classes names.
        :rtype: list(str)

        Examples
        --------
        >>> from gemseo.caches.cache_factory import CacheFactory
        >>> CacheFactory().caches
        ['AbstractFullCache', 'HDF5Cache', 'MemoryFullCache', 'SimpleCache']
        """
        return self.factory.classes

    def is_available(self, cache_name):
        """Checks the availability of a cache.

        :param str cache_name:  cache_name of the cache.
        :returns: True if the cache is available.
        :rtype: bool

        Examples
        --------
        >>> from gemseo.caches.cache_factory import CacheFactory
        >>> CacheFactory().is_available('SimpleCache')
        True
        >>> CacheFactory().is_available('UnavailableCache')
        False
        """
        return self.factory.is_available(cache_name)
